#!/usr/bin/env python3

__author__ = 'Frederic Escudie'
__copyright__ = 'Copyright (C) 2020 IUCT-O'
__license__ = 'GNU General Public License'
__version__ = '1.0.0'
__email__ = 'escudie.frederic@iuct-oncopole.fr'
__status__ = 'prod'

import os
import sys
import tempfile
import unittest
import uuid

TEST_DIR = os.path.dirname(os.path.abspath(__file__))
PACKAGE_DIR = os.path.dirname(TEST_DIR)
sys.path.append(PACKAGE_DIR)

from anacore.picardIO import castCol, getColType, getValType, PicardReader


########################################################################
#
# FUNCTIONS
#
########################################################################
class TestCastCol(unittest.TestCase):
    def testCastCol(self):
        # To str
        data = ["a", "b", "c", "10", "", "15.1", "16.3"]
        expected = ["a", "b", "c", "10", None, "15.1", "16.3"]
        castCol(data, "str")
        self.assertEqual(expected, data)
        # To int
        data = ["10", "", "15", "16"]
        expected = [10, None, 15, 16]
        castCol(data, "int")
        self.assertEqual(expected, data)
        # To float
        data = ["10", "15.1", "", "16.3"]
        expected = [10.0, 15.1, None, 16.3]
        castCol(data, "float")
        self.assertEqual(expected, data)


class TestGetColType(unittest.TestCase):
    def testGetColType(self):
        # Is None
        data = [None, ""]
        expected = None
        observed = getColType(data)
        self.assertEqual(expected, observed)
        # Is str
        data = ["a", "b", "c", "10", None, "", "15.1", "16.3"]
        expected = "str"
        observed = getColType(data)
        self.assertEqual(expected, observed)
        # Is int
        data = ["10", None, "", "15", "16"]
        expected = "int"
        observed = getColType(data)
        self.assertEqual(expected, observed)
        # Is float
        data = ["10", None, "", "15.1", "16"]
        expected = "float"
        observed = getColType(data)
        self.assertEqual(expected, observed)


class TestGetValType(unittest.TestCase):
    def testGetValType(self):
        data = ["a", "b", "c", "10", None, "", "15.1", "16.3"]
        expected = ["str", "str", "str", "int", None, "str", "float", "float"]
        observed = [getValType(elt) for elt in data]
        self.assertEqual(expected, observed)


class TestPicardReader(unittest.TestCase):
    def setUp(self):
        tmp_folder = tempfile.gettempdir()
        unique_id = str(uuid.uuid1())
        # CollectArraysVariantCallingMetrics
        self.tmp_variants_call = os.path.join(tmp_folder, unique_id + "_vcm.txt")
        with open(self.tmp_variants_call, "w") as writer:
            writer.write("""## htsjdk.samtools.metrics.StringHeader
# CollectArraysVariantCallingMetrics  --INPUT /cromwell_root/broad-gotc-dev-cromwell-execution/Arrays/9972b43c-df1e-414a-9017-24edc37b5250/call-IlluminaGenotypingArray/IlluminaGenotypingArray/805b2e59-add2-4794-ae89-350d69a0f7c3/call-MergePedIntoVcf/7991775143_R01C01.vcf.gz --OUTPUT 7991775143_R01C01 --CALL_RATE_PF_THRESHOLD 0.98 --DBSNP /cromwell_root/gcp-public-data--broad-references/hg19/v0/dbsnp_138.b37.vcf.gz  --NUM_PROCESSORS 0 --VERBOSITY INFO --QUIET false --VALIDATION_STRINGENCY STRICT --COMPRESSION_LEVEL 5 --MAX_RECORDS_IN_RAM 500000 --CREATE_INDEX false --CREATE_MD5_FILE false --GA4GH_CLIENT_SECRETS client_secrets.json --help false --version false --showHidden false --USE_JDK_DEFLATER false --USE_JDK_INFLATER false
## htsjdk.samtools.metrics.StringHeader
# Started on: Thu Mar 05 18:21:35 UTC 2020

## METRICS CLASS	picard.arrays.CollectArraysVariantCallingMetrics$ArraysVariantCallingDetailMetrics
CHIP_WELL_BARCODE	SAMPLE_ALIAS	ANALYSIS_VERSION	CHIP_TYPE	AUTOCALL_PF	AUTOCALL_DATE	IMAGING_DATE	IS_ZCALLED	AUTOCALL_GENDER	FP_GENDER	REPORTED_GENDER	GENDER_CONCORDANCE_PF	HET_PCT	CLUSTER_FILE_NAME	P95_GREEN	P95_RED	AUTOCALL_VERSION	ZCALL_VERSION	EXTENDED_MANIFEST_VERSION	HET_HOMVAR_RATIO	SCANNER_NAME	NUM_ASSAYS	NUM_NON_FILTERED_ASSAYS	NUM_FILTERED_ASSAYS	NUM_ZEROED_OUT_ASSAYS	NUM_SNPS	NUM_INDELS	NUM_CALLS	NUM_AUTOCALL_CALLS	NUM_NO_CALLS	NUM_IN_DB_SNP	NOVEL_SNPS	PCT_DBSNP	CALL_RATE	AUTOCALL_CALL_RATE	NUM_SINGLETONS
7991775143_R01C01	NA12878	1	HumanExome-12v1-1_A	Y	2019-12-14T03:05:00+0000	2012-12-16T16:10:10+0000	Y	F	U	F	Y	0.042993	HumanExomev1_1_CEPH_A.egt	4251	5479	3.0.0	1.0.0.0	1.3	1.525553	N370	242513	242493	20	0	242450	43	242318	241250	175	238085	4365	0.981996	0.999278	0.994874	10418

""")
        # CollectWgsMetrics
        self.tmp_wgs = os.path.join(tmp_folder, unique_id + "_wgsm.txt")
        with open(self.tmp_wgs, "w") as writer:
            writer.write("""## htsjdk.samtools.metrics.StringHeader
# CollectWgsMetrics INPUT=/cromwell_root/broad-gotc-dev-cromwell-execution/WholeGenomeGermlineSingleSample/e36be5d3-ddbf-4ee6-a98d-2a1933826550/call-UnmappedBamToAlignedBam/UnmappedBamToAlignedBam/da7029e5-cf73-4d9b-8369-23c3aea72f19/call-GatherBamFiles/NA12878_PLUMBING.bam OUTPUT=NA12878_PLUMBING.wgs_metrics INCLUDE_BQ_HISTOGRAM=true INTERVALS=/cromwell_root/gcp-public-data--broad-references/hg38/v0/wgs_coverage_regions.hg38.interval_list USE_FAST_ALGORITHM=true READ_LENGTH=250 VALIDATION_STRINGENCY=SILENT REFERENCE_SEQUENCE=/cromwell_root/gcp-public-data--broad-references/hg38/v0/Homo_sapiens_assembly38.fasta    MINIMUM_MAPPING_QUALITY=20 MINIMUM_BASE_QUALITY=20 COVERAGE_CAP=250 LOCUS_ACCUMULATION_CAP=100000 STOP_AFTER=-1 COUNT_UNPAIRED=false SAMPLE_SIZE=10000 ALLELE_FRACTION=[0.001, 0.005, 0.01, 0.02, 0.05, 0.1, 0.2, 0.3, 0.5] VERBOSITY=INFO QUIET=false COMPRESSION_LEVEL=5 MAX_RECORDS_IN_RAM=500000 CREATE_INDEX=false CREATE_MD5_FILE=false GA4GH_CLIENT_SECRETS=client_secrets.json USE_JDK_DEFLATER=false USE_JDK_INFLATER=false
## htsjdk.samtools.metrics.StringHeader
# Started on: Fri Mar 20 20:49:11 UTC 2020

## METRICS CLASS	picard.analysis.WgsMetrics
GENOME_TERRITORY	MEAN_COVERAGE	SD_COVERAGE	MEDIAN_COVERAGE	MAD_COVERAGE	PCT_EXC_ADAPTER	PCT_EXC_MAPQ	PCT_EXC_DUPE	PCT_EXC_UNPAIRED	PCT_EXC_BASEQ	PCT_EXC_OVERLAP	PCT_EXC_CAPPED	PCT_EXC_TOTAL	PCT_1X	PCT_5X	PCT_10X	PCT_15X	PCT_20X	PCT_25X	PCT_30X	PCT_40X	PCT_50X	PCT_60X	PCT_70X	PCT_80X	PCT_90X	PCT_100X	HET_SNP_SENSITIVITY	HET_SNP_Q
2745186692	0.003328	0.059988	0	0	0.000045	0.044363	0.000482	0.00674	0.121016	0.090623	0	0.263269	0.003305	0.000001	0	0	0	0	0	0	0	0	0	0	0	0	0.000105	0

## HISTOGRAM	java.lang.Integer
coverage	high_quality_coverage_count	unfiltered_baseq_count
0	2736113146	0
1	9041943	0
2	24393	0
3	3102	730
4	1433	940
5	779	10184
6	861	1526
7	236	0
8	149	0
9	83	0
10	75	45052
11	54	0
12	0	0
13	0	0
14	0	0
15	0	0
16	0	0
17	1	0
18	1	0
19	10	0
20	0	67107
21	2	0
22	0	0
23	3	0
24	11	0
25	7	0
26	16	0
27	1	0
28	13	0
29	17	0
30	23	10192669
31	12	0
32	22	0
""")
        # CollectInsertSizeMetrics
        self.tmp_insert = os.path.join(tmp_folder, unique_id + "_insert.txt")
        with open(self.tmp_insert, "w") as writer:
            writer.write("""## htsjdk.samtools.metrics.StringHeader
# picard.analysis.CollectInsertSizeMetrics INPUT=/path/to/bam REFERENCE_SEQUENCE=/path/to/reference ASSUME_SORTED=true OUTPUT=/path/to/output H=/path/to/histogram STOP_AFTER=0 VERBOSITY=INFO QUIET=false VALIDATION_STRINGENCY=STRICT COMPRESSION_LEVEL=5 MAX_RECORDS_IN_RAM=500000 CREATE_INDEX=false CREATE_MD5_FILE=false
## htsjdk.samtools.metrics.StringHeader
# Started on: Fri Nov 13 20:25:43 CET 2020

## METRICS CLASS	picard.analysis.InsertSizeMetrics
MEDIAN_INSERT_SIZE	MODE_INSERT_SIZE	MEDIAN_ABSOLUTE_DEVIATION	MIN_INSERT_SIZE	MAX_INSERT_SIZE	MEAN_INSERT_SIZE	STANDARD_DEVIATION	READ_PAIRS	PAIR_ORIENTATION	WIDTH_OF_10_PERCENT	WIDTH_OF_20_PERCENT	WIDTH_OF_30_PERCENT	WIDTH_OF_40_PERCENT	WIDTH_OF_50_PERCENT	WIDTH_OF_60_PERCENT	WIDTH_OF_70_PERCENT	WIDTH_OF_80_PERCENT	WIDTH_OF_90_PERCENT	WIDTH_OF_95_PERCENT	WIDTH_OF_99_PERCENT	SAMPLE	LIBRARY	READ_GROUP
286	166	70	14	244976752	260.992864	87.727108	14114058	FR	23	43	67	99	141	193	227	243	269	299	437
98	50	1	96	105	1.1	1.1	1	RF	1	1	1	1	1	1	1	1	1	1	1
98	40	1	97	98	1.1	1.1	1	TANDEM	1	1	1	1	1	1	1	1	1	1	1

## HISTOGRAM	java.lang.Integer
insert_size	All_Reads.fr_count	All_Reads.rf_count	All_Reads.tandem_count
1	0	100	0
2	0	100	0
3	0	100	0
4	99	100	0
5	0	100	0
6	0	100	0
7	0	100	0
8	0	100	0
9	0	100	0
10	0	100	0
11	0	100	0
12	0	100	0
13	0	100	50
14	0	100	0
15	0	100	0
16	0	100	0
17	0	100	0
18	0	100	0
19	0	10	0
20	0	100	0""")

    def tearDown(self):
        # Clean temporary files
        for curr_file in [self.tmp_variants_call, self.tmp_wgs, self.tmp_insert]:
            if os.path.exists(curr_file):
                os.remove(curr_file)

    def testCollectInsertSizeMetrics(self):
        observed = PicardReader(self.tmp_insert)
        expected = {
            "command": "picard.analysis.CollectInsertSizeMetrics INPUT=/path/to/bam REFERENCE_SEQUENCE=/path/to/reference ASSUME_SORTED=true OUTPUT=/path/to/output H=/path/to/histogram STOP_AFTER=0 VERBOSITY=INFO QUIET=false VALIDATION_STRINGENCY=STRICT COMPRESSION_LEVEL=5 MAX_RECORDS_IN_RAM=500000 CREATE_INDEX=false CREATE_MD5_FILE=false",
            "metrics": [
                {
                    'MEDIAN_INSERT_SIZE': 286,
                    'MODE_INSERT_SIZE': 166,
                    'MEDIAN_ABSOLUTE_DEVIATION': 70,
                    'MIN_INSERT_SIZE': 14,
                    'MAX_INSERT_SIZE': 244976752,
                    'MEAN_INSERT_SIZE': 260.992864,
                    'STANDARD_DEVIATION': 87.727108,
                    'READ_PAIRS': 14114058,
                    'PAIR_ORIENTATION': 'FR',
                    'WIDTH_OF_10_PERCENT': 23,
                    'WIDTH_OF_20_PERCENT': 43,
                    'WIDTH_OF_30_PERCENT': 67,
                    'WIDTH_OF_40_PERCENT': 99,
                    'WIDTH_OF_50_PERCENT': 141,
                    'WIDTH_OF_60_PERCENT': 193,
                    'WIDTH_OF_70_PERCENT': 227,
                    'WIDTH_OF_80_PERCENT': 243,
                    'WIDTH_OF_90_PERCENT': 269,
                    'WIDTH_OF_95_PERCENT': 299,
                    'WIDTH_OF_99_PERCENT': 437,
                    'SAMPLE': None,
                    'LIBRARY': None,
                    'READ_GROUP': None
                }, {
                    'MEDIAN_INSERT_SIZE': 98, 'MODE_INSERT_SIZE': 50, 'MEDIAN_ABSOLUTE_DEVIATION': 1, 'MIN_INSERT_SIZE': 96, 'MAX_INSERT_SIZE': 105, 'MEAN_INSERT_SIZE': 1.1, 'STANDARD_DEVIATION': 1.1, 'READ_PAIRS': 1, 'PAIR_ORIENTATION': 'RF', 'WIDTH_OF_10_PERCENT': 1, 'WIDTH_OF_20_PERCENT': 1, 'WIDTH_OF_30_PERCENT': 1, 'WIDTH_OF_40_PERCENT': 1, 'WIDTH_OF_50_PERCENT': 1, 'WIDTH_OF_60_PERCENT': 1, 'WIDTH_OF_70_PERCENT': 1, 'WIDTH_OF_80_PERCENT': 1, 'WIDTH_OF_90_PERCENT': 1, 'WIDTH_OF_95_PERCENT': 1, 'WIDTH_OF_99_PERCENT': 1, 'SAMPLE': None, 'LIBRARY': None, 'READ_GROUP': None
                }, {
                    'MEDIAN_INSERT_SIZE': 98, 'MODE_INSERT_SIZE': 40, 'MEDIAN_ABSOLUTE_DEVIATION': 1, 'MIN_INSERT_SIZE': 97, 'MAX_INSERT_SIZE': 98, 'MEAN_INSERT_SIZE': 1.1, 'STANDARD_DEVIATION': 1.1, 'READ_PAIRS': 1, 'PAIR_ORIENTATION': 'TANDEM', 'WIDTH_OF_10_PERCENT': 1, 'WIDTH_OF_20_PERCENT': 1, 'WIDTH_OF_30_PERCENT': 1, 'WIDTH_OF_40_PERCENT': 1, 'WIDTH_OF_50_PERCENT': 1, 'WIDTH_OF_60_PERCENT': 1, 'WIDTH_OF_70_PERCENT': 1, 'WIDTH_OF_80_PERCENT': 1, 'WIDTH_OF_90_PERCENT': 1, 'WIDTH_OF_95_PERCENT': 1, 'WIDTH_OF_99_PERCENT': 1, 'SAMPLE': None, 'LIBRARY': None, 'READ_GROUP': None
                }
            ],
            "histogram": {
                'insert_size': [1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20],
                'All_Reads.fr_count': [0, 0, 0, 99, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
                'All_Reads.rf_count': [100, 100, 100, 100, 100, 100, 100, 100, 100, 100, 100, 100, 100, 100, 100, 100, 100, 100, 10, 100],
                'All_Reads.tandem_count': [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 50, 0, 0, 0, 0, 0, 0, 0]
            }
        }
        self.assertEqual(expected["command"], observed.command)
        self.assertEqual(expected["metrics"], observed.metrics)
        self.assertEqual(expected["histogram"], observed.histogram)

    def testCollectArraysVariantCallingMetrics(self):
        observed = PicardReader(self.tmp_variants_call)
        expected = {
            "command": "CollectArraysVariantCallingMetrics  --INPUT /cromwell_root/broad-gotc-dev-cromwell-execution/Arrays/9972b43c-df1e-414a-9017-24edc37b5250/call-IlluminaGenotypingArray/IlluminaGenotypingArray/805b2e59-add2-4794-ae89-350d69a0f7c3/call-MergePedIntoVcf/7991775143_R01C01.vcf.gz --OUTPUT 7991775143_R01C01 --CALL_RATE_PF_THRESHOLD 0.98 --DBSNP /cromwell_root/gcp-public-data--broad-references/hg19/v0/dbsnp_138.b37.vcf.gz  --NUM_PROCESSORS 0 --VERBOSITY INFO --QUIET false --VALIDATION_STRINGENCY STRICT --COMPRESSION_LEVEL 5 --MAX_RECORDS_IN_RAM 500000 --CREATE_INDEX false --CREATE_MD5_FILE false --GA4GH_CLIENT_SECRETS client_secrets.json --help false --version false --showHidden false --USE_JDK_DEFLATER false --USE_JDK_INFLATER false",
            "metrics": [
                {
                    'CHIP_WELL_BARCODE': '7991775143_R01C01',
                    'SAMPLE_ALIAS': 'NA12878',
                    'ANALYSIS_VERSION': 1,
                    'CHIP_TYPE': 'HumanExome-12v1-1_A',
                    'AUTOCALL_PF': 'Y',
                    'AUTOCALL_DATE': '2019-12-14T03:05:00+0000',
                    'IMAGING_DATE': '2012-12-16T16:10:10+0000',
                    'IS_ZCALLED': 'Y',
                    'AUTOCALL_GENDER': 'F',
                    'FP_GENDER': 'U',
                    'REPORTED_GENDER': 'F',
                    'GENDER_CONCORDANCE_PF': 'Y',
                    'HET_PCT': 0.042993,
                    'CLUSTER_FILE_NAME': 'HumanExomev1_1_CEPH_A.egt',
                    'P95_GREEN': 4251,
                    'P95_RED': 5479,
                    'AUTOCALL_VERSION': '3.0.0',
                    'ZCALL_VERSION': '1.0.0.0',
                    'EXTENDED_MANIFEST_VERSION': 1.3,
                    'HET_HOMVAR_RATIO': 1.525553,
                    'SCANNER_NAME': 'N370',
                    'NUM_ASSAYS': 242513,
                    'NUM_NON_FILTERED_ASSAYS': 242493,
                    'NUM_FILTERED_ASSAYS': 20,
                    'NUM_ZEROED_OUT_ASSAYS': 0,
                    'NUM_SNPS': 242450,
                    'NUM_INDELS': 43, 'NUM_CALLS': 242318,
                    'NUM_AUTOCALL_CALLS': 241250,
                    'NUM_NO_CALLS': 175,
                    'NUM_IN_DB_SNP': 238085,
                    'NOVEL_SNPS': 4365,
                    'PCT_DBSNP': 0.981996,
                    'CALL_RATE': 0.999278,
                    'AUTOCALL_CALL_RATE': 0.994874,
                    'NUM_SINGLETONS': 10418
                }
            ],
            "histogram": None
        }
        self.assertEqual(expected["command"], observed.command)
        self.assertEqual(expected["metrics"], observed.metrics)
        self.assertEqual(expected["histogram"], observed.histogram)

    def testCollectWgsMetrics(self):
        observed = PicardReader(self.tmp_wgs)
        expected = {
            "command": "CollectWgsMetrics INPUT=/cromwell_root/broad-gotc-dev-cromwell-execution/WholeGenomeGermlineSingleSample/e36be5d3-ddbf-4ee6-a98d-2a1933826550/call-UnmappedBamToAlignedBam/UnmappedBamToAlignedBam/da7029e5-cf73-4d9b-8369-23c3aea72f19/call-GatherBamFiles/NA12878_PLUMBING.bam OUTPUT=NA12878_PLUMBING.wgs_metrics INCLUDE_BQ_HISTOGRAM=true INTERVALS=/cromwell_root/gcp-public-data--broad-references/hg38/v0/wgs_coverage_regions.hg38.interval_list USE_FAST_ALGORITHM=true READ_LENGTH=250 VALIDATION_STRINGENCY=SILENT REFERENCE_SEQUENCE=/cromwell_root/gcp-public-data--broad-references/hg38/v0/Homo_sapiens_assembly38.fasta    MINIMUM_MAPPING_QUALITY=20 MINIMUM_BASE_QUALITY=20 COVERAGE_CAP=250 LOCUS_ACCUMULATION_CAP=100000 STOP_AFTER=-1 COUNT_UNPAIRED=false SAMPLE_SIZE=10000 ALLELE_FRACTION=[0.001, 0.005, 0.01, 0.02, 0.05, 0.1, 0.2, 0.3, 0.5] VERBOSITY=INFO QUIET=false COMPRESSION_LEVEL=5 MAX_RECORDS_IN_RAM=500000 CREATE_INDEX=false CREATE_MD5_FILE=false GA4GH_CLIENT_SECRETS=client_secrets.json USE_JDK_DEFLATER=false USE_JDK_INFLATER=false",
            "metrics": [{'GENOME_TERRITORY': 2745186692, 'MEAN_COVERAGE': 0.003328, 'SD_COVERAGE': 0.059988, 'MEDIAN_COVERAGE': 0, 'MAD_COVERAGE': 0, 'PCT_EXC_ADAPTER': 4.5e-05, 'PCT_EXC_MAPQ': 0.044363, 'PCT_EXC_DUPE': 0.000482, 'PCT_EXC_UNPAIRED': 0.00674, 'PCT_EXC_BASEQ': 0.121016, 'PCT_EXC_OVERLAP': 0.090623, 'PCT_EXC_CAPPED': 0, 'PCT_EXC_TOTAL': 0.263269, 'PCT_1X': 0.003305, 'PCT_5X': 1e-06, 'PCT_10X': 0, 'PCT_15X': 0, 'PCT_20X': 0, 'PCT_25X': 0, 'PCT_30X': 0, 'PCT_40X': 0, 'PCT_50X': 0, 'PCT_60X': 0, 'PCT_70X': 0, 'PCT_80X': 0, 'PCT_90X': 0, 'PCT_100X': 0, 'HET_SNP_SENSITIVITY': 0.000105, 'HET_SNP_Q': 0}],
            "histogram": {
                'coverage': [0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32],
                'high_quality_coverage_count': [2736113146, 9041943, 24393, 3102, 1433, 779, 861, 236, 149, 83, 75, 54, 0, 0, 0, 0, 0, 1, 1, 10, 0, 2, 0, 3, 11, 7, 16, 1, 13, 17, 23, 12, 22],
                'unfiltered_baseq_count': [0, 0, 0, 730, 940, 10184, 1526, 0, 0, 0, 45052, 0, 0, 0, 0, 0, 0, 0, 0, 0, 67107, 0, 0, 0, 0, 0, 0, 0, 0, 0, 10192669, 0, 0]
            }
        }
        self.assertEqual(expected["command"], observed.command)
        self.assertEqual(expected["metrics"], observed.metrics)
        self.assertEqual(expected["histogram"], observed.histogram)


########################################################################
#
# MAIN
#
########################################################################
if __name__ == "__main__":
    unittest.main()



# -*- coding: utf-8 -*-
"""Classes and functions for reading/writing Separated Values files (CSV, TSV, ...)."""

__author__ = 'Frederic Escudie'
__copyright__ = 'Copyright (C) 2017 IUCT-O'
__license__ = 'GNU General Public License'
__version__ = '1.1.0'
__email__ = 'escudie.frederic@iuct-oncopole.fr'
__status__ = 'prod'

from anacore.abstractFile import isEmpty, AbstractFile


class SVIO(AbstractFile):
    """Class to read and write in separated value files like TSV, CSV, etc. Each rows is view as a list."""

    def __init__(self, filepath, mode="r", separator="\t", title_starter=None, has_title=True, metadata_starter="##"):
        """
        Build and return an instance of SVIO.

        :param filepath: The filepath.
        :type filepath: str.
        :param mode: Mode to open the file ('r', 'w', 'a').
        :type mode: str.
        :param separator: Separator used between values.
        :type separator: str.
        :param title_starter: The string used to introduce the title line.
        :type title_starter: str.
        :param has_title: If the first line contains the titles of columns.
        :type has_title: bool.
        :param title_metadata: The string used to introduce a metadata line.
        :type title_metadata: str.
        :return: The new instance.
        :rtype: SVIO
        """
        # Convert mode for append in empty file
        if mode == "a" and isEmpty(filepath):
            mode = "w"
        # Get existing titles and metadata if the file is in append mode
        pre_nb_line = 0
        pre_titles = None
        pre_metadata = []
        if mode == "a":  # Get titles from existing file
            with SVIO(filepath, "r", separator, title_starter, has_title) as reader:
                pre_titles = reader.titles
                pre_metadata = reader.metadata
                pre_nb_line = reader.current_line_nb + sum([1 for line in reader.file_handle])
        # Initialise instance
        AbstractFile.__init__(self, filepath, mode)
        self.separator = separator
        self.metadata = pre_metadata
        self.metadata_starter = metadata_starter
        self._has_title = has_title
        self.titles = pre_titles
        self.title_starter = title_starter
        if mode == "r":
            self._parseHeader()
        elif mode == "a":
            self.current_line_nb = pre_nb_line

    def _parseHeader(self):
        """Parse SV header to set the attributes titles and metadata."""
        if self.current_line_nb == 0 and not isEmpty(self.filepath):
            # Metadata
            curr_line_pointer = self.file_handle.tell()
            self.current_line = self.file_handle.readline().rstrip()
            self.current_line_nb += 1
            while self.current_line.startswith(self.metadata_starter):
                self.metadata.append(self.current_line[len(self.metadata_starter):])
                curr_line_pointer = self.file_handle.tell()
                self.current_line = self.file_handle.readline().rstrip()
                self.current_line_nb += 1
            # Titles
            if self._has_title:
                cleaned_line = self.current_line
                if self.title_starter is not None and self.title_starter != "":
                    if not self.current_line.startswith(self.title_starter):
                        raise Exception('The title line does not starts with "{}".'.format(self.title_starter))
                    cleaned_line = self.current_line[len(self.title_starter):]
                self.titles = [elt.strip() for elt in cleaned_line.split(self.separator)]
            else:
                self.file_handle.seek(curr_line_pointer)
                self.current_line_nb -= 1

    def isRecordLine(self, line):
        """
        Return True if the line corresponds to a record (it is not a comment or an header line).

        :param line: The evaluated line.
        :type line: str.
        :return: True if the line corresponds to a record.
        :rtype: bool
        """
        is_record = True
        if line.startswith("#"):
            is_record = False
        return is_record

    def _parseLine(self):
        """
        Return a structured record from the current line.

        :return: The record defined by the current line.
        :rtype: list
        """
        fields = [field.strip() for field in self.current_line.split(self.separator)]
        return fields

    @staticmethod
    def isValid(filepath, separator="\t", metadata_starter="##"):
        """
        Return True is the file can be a SV file.

        :return: True if the file can be a SV file.
        :rtype: bool
        """
        is_valid = False
        pointer_after_metadata = False
        nb_fields = list()  # The number of fields for each row
        reader = SVIO(filepath, separator=separator, title_starter="", has_title=False, metadata_starter="##")
        try:
            # Read the 10 first lines
            line_idx = 1
            line = reader.file_handle.readline()
            if line is None:  # File is empty
                is_valid = True
            else:  # File has content
                if not line.startswith(metadata_starter):
                    pointer_after_metadata = True
                    nb_fields.append(line.count(separator))
                while line_idx < 10 and line:
                    line = reader.file_handle.readline()
                    if line:
                        if pointer_after_metadata or not line.startswith(metadata_starter):
                            pointer_after_metadata = True
                            nb_fields.append(line.count(separator))
                            line_idx += 1
                if len(set(nb_fields)) == 1 and nb_fields[0] > 1:  # All lines have the same number of columns
                    is_valid = True
        except Exception:
            pass
        finally:
            reader.close()
        return is_valid

    def writeHeader(self):
        """Write header lines in file."""
        # Metadata
        for curr_metadata in self.metadata:
            self.file_handle.write(
                self.metadata_starter + curr_metadata + "\n"
            )
            self.current_line_nb += 1
        # Titles
        if self.titles is not None:
            line_starter = "" if self.title_starter is None else self.title_starter
            self.file_handle.write(
                line_starter + self.separator.join(self.titles) + "\n"
            )
            self.current_line_nb += 1

    def write(self, record):
        """
        Write record line in file.

        :param record: The record.
        :type record: list.
        """
        # Write header if the file is empty
        if self.current_line_nb == 0:
            self.writeHeader()
        # Write record
        self.file_handle.write(self.recordToLine(record) + "\n")
        self.current_line_nb += 1

    def recordToLine(self, record):
        """
        Return the record in SV format.

        :param record: The record to process.
        :type record: list.
        :return: The SV line corresponding to the record.
        :rtype: str
        """
        line = self.separator.join([str(elt) for elt in record])
        return(line)


class HashedSVIO(SVIO):
    """Class to read and write in separated value files like TSV, CSV, etc. Each rows is view as a dict indexed by columns titles."""

    def __init__(self, filepath, mode="r", separator="\t", title_starter=None, metadata_starter="##"):
        """
        Build and return an instance of HashedSVIO.

        :param filepath: The filepath.
        :type filepath: str.
        :param mode: Mode to open the file ('r', 'w', 'a').
        :type mode: str.
        :param separator: Separator used between values.
        :type separator: str.
        :param title_starter: The string used to introduce the title line.
        :type title_starter: str.
        :param title_metadata: The string used to introduce a metadata line.
        :type title_metadata: str.
        :return: The new instance.
        :rtype: HashedSVIO
        """
        super().__init__(filepath, mode, separator, title_starter, True, metadata_starter)

    def _parseLine(self):
        """
        Return a structured record from the current line.

        :return: The record defined by the current line.
        :rtype: dict
        """
        fields = super()._parseLine()
        record = {self.titles[idx]: val for idx, val in enumerate(fields)}
        return record

    def recordToLine(self, record):
        """
        Return the record in SV format.

        :param record: The record to process.
        :type record: dict.
        :return: The SV line corresponding to the record.
        :rtype: str
        """
        fields = [record[curr_title] for curr_title in self.titles]
        line = self.separator.join([str(elt) for elt in fields])
        return(line)



# -*- coding: utf-8 -*-
"""
Classes and functions for reading/writing fusions callers's output and VCF containing fusions. Each record is converted to VCFRecord.

:Example:

    Read results form STAR-Fusion

    .. highlight:: python
    .. code-block:: python

        # File content>
        # #FusionName	JunctionReadCount	SpanningFragCount	SpliceType	LeftGene	LeftBreakpoint	RightGene	RightBreakpoint	LargeAnchorSupport	FFPM	LeftBreakDinuc	LeftBreakEntropy	RightBreakDinuc	RightBreakEntropy	annots
        # MN1--BEND2	90	39	ONLY_REF_SPLICE	MN1^ENSG00000169184.6	chr22:27796763:-	BEND2^ENSG00000177324.14	chrX:18195442:-	YES_LDAS	38.9246	GT	1.9656	AG	1.7232	["INTERCHROMOSOMAL[chr22--chr9]"]

        from anacore.fusion import FusionFileReader

        spl_name = "spl1"
        print("Breakpoint_1\\tGene_brkpt_1\\tBreakpoint_2\\tGene_brkpt_2\\tSpanning_pair\\tSplit_read")
        with FusionFileReader.factory("test.tsv", sample_name=spl_name) as reader:
            for first, second in reader:
                print(
                    "{}:{}".format(first.chrom, frist.pos),
                    ",".join([elt["SYMBOL"] for elt in first.info["ANN"]]),
                    "{}:{}".format(second.chrom, second.pos),
                    ",".join([elt["SYMBOL"] for elt in second.info["ANN"]]),
                    first.sample["spl_name"]["PR"],
                    first.sample["spl_name"]["SR"],
                    sep="\\t"
                )

        # Result>
        # Breakpoint_1\tGene_brkpt_1\tBreakpoint_2\tGene_brkpt_2\tSpanning_pair\tSplit_read
        # chr22:27796763\tMN1\tchrX:18195442\tBEND2\t39\t99

    Read VCF

    .. highlight:: python
    .. code-block:: python

        from anacore.fusion import FusionFileReader

        # File content>
        # ##fileformat=VCFv4.3
        # ##INFO=<ID=MATEID,Number=A,Type=String,Description="ID of mate breakend.">
        # ##INFO=<ID=SVTYPE,Number=1,Type=String,Description="Type of structural variant.">
        # #CHROM	POS	ID	REF	ALT	QUAL	FILTER	INFO
        # 2	321682	bnd_V	T	]13:123456]T	6	PASS	SVTYPE=BND;MATEID=bnd_U
        # 4	888888	.	T	G	.	PASS	.
        # 13	123456	bnd_U	C	C[2:321682[,C[17:198983[	6	PASS	SVTYPE=BND;MATEID=bnd_V,bnd_Z
        # 17	198983	bnd_Z	A	]13:123456]A	6	PASS	SVTYPE=BND;MATEID=bnd_U

        print("Breakpoint_1\\tBreakpoint_1_name\\tBreakpoint_2\\tBreakpoint_2_name")
        with FusionFileReader.factory("test.vcf") as reader:
            for first, second in reader:
                print(
                    "{}:{}".format(first.chrom, frist.pos),
                    first.id
                    "{}:{}".format(second.chrom, second.pos),
                    second.id
                    sep="\\t"
                )

        # Result>
        # Breakpoint_1\tBreakpoint_1_name\tBreakpoint_2\tBreakpoint_2_name
        # 2:321682\tbnd_V\t13:123456\tbnd_U_0
        # 13:123456\tbnd_U_1\t17:198983\tbnd_Z
"""
"""
Alt representation in VCF specification:

    First in RNA and strand + => alt: N|
                    ------>
                           |
        >>>>>>>>>>>>>>     |     >>>>>>>>>>>>
        ...................|.................

                    ------>
                           |
        <<<<<<<<<<<<<<     |     <<<<<<<<<<<<
        ...................|.................

    Second in RNA and strand + => alt: |N
                            ------>
                           |
        >>>>>>>>>>>>>>     |     >>>>>>>>>>>>
        ...................|.................

                            ------>
                           |
        <<<<<<<<<<<<<<     |     <<<<<<<<<<<<
        ...................|.................

    First in RNA and strand - => alt: |N
                            <------
                           |
        >>>>>>>>>>>>>>     |     >>>>>>>>>>>>
        ...................|.................

                            <------
                           |
        <<<<<<<<<<<<<<     |     <<<<<<<<<<<<
        ...................|.................

    Second in RNA and strand - => alt: N|
                    <------
                           |
        >>>>>>>>>>>>>>     |     >>>>>>>>>>>>
        ...................|.................

                    <------
                           |
        <<<<<<<<<<<<<<     |     <<<<<<<<<<<<
        ...................|.................
"""

__author__ = 'Frederic Escudie'
__copyright__ = 'Copyright (C) 2019 IUCT-O'
__license__ = 'GNU General Public License'
__version__ = '2.6.1'
__email__ = 'escudie.frederic@iuct-oncopole.fr'
__status__ = 'prod'

import gzip
import json
import re
import uuid
from anacore.abstractFile import isGzip
from anacore.annotVcf import AnnotVCFIO
from anacore.sv import HashedSVIO, SVIO
from anacore.vcf import decodeInfoValue, getAlleleRecord, VCFIO, VCFRecord, HeaderInfoAttr, HeaderFilterAttr, HeaderFormatAttr


def getBNDInterval(record):
    """
    Return the start and end for BND record (CIPOS is taken into account).

    :param record: The breakend record.
    :type record: anacore.vcf.VCFRecord
    :return: Start and end for BND record (CIPOS is taken into account).
    :rtype: (int, int)
    """
    start = record.pos
    end = start
    if "CIPOS" in record.info:
        start += record.info["CIPOS"][0]
        end += record.info["CIPOS"][1]
        if "IMPRECISE" in record.info:
            start -= 1
    return start, end


def getCoordDictFromCoordStr(coord):
    """
    Return elements from coordinates defined by a fusion caller ("CHROM:POS:STRAND").

    :param coord: Coordinates defined by fusion caller in format "CHROM:POS:STRAND".
    :type coord: str
    :return: Coordinates in dict format {"chrom": "CHROM", "pos": "POS", "strand":"STRAND"}.
    :rtype: dict
    """
    matches = re.match("(.+):(\d+):([+-.])", coord)
    if matches is None:
        raise Exception("The coordinates fields {} cannot be parsed.".format(coord))
    return {
        "chrom": matches.group(1),
        "pos": int(matches.group(2)),
        "strand": matches.group(3)
    }


def getAltFromCoord(first_coord, second_coord):
    """
    Return VCF alternative field value for the breakends from the 5' and 3' shard of fusion.

    :param first_coord: Coordinates of the breakend for the first 5' shard in fusion.
    :type first_coord: dict
    :param second_coord: Coordinates of the breakend for the first 3' shard in fusion.
    :type second_coord: dict
    :return: Alternative for the first shard in fusion and alternative for the second shard in fusion.
    :rtype: (str, str)
    """
    first_alt = None
    second_alt = None
    if first_coord["strand"] == "." or second_coord["strand"] == ".":
        raise Exception("getAltFromCoord cannot process coordinates with unknown strand.")
    if first_coord["strand"] == "+":
        if second_coord["strand"] == "-":  # First shard is + and second is -
            first_alt = "N]{}:{}]".format(second_coord["chrom"], second_coord["pos"])
            second_alt = "N]{}:{}]".format(first_coord["chrom"], first_coord["pos"])
        else:   # First shard is + and second is +
            first_alt = "N[{}:{}[".format(second_coord["chrom"], second_coord["pos"])
            second_alt = "]{}:{}]N".format(first_coord["chrom"], first_coord["pos"])
    else:
        if second_coord["strand"] == "-":  # First shard is - and second is -
            first_alt = "]{}:{}]N".format(second_coord["chrom"], second_coord["pos"])
            second_alt = "N[{}:{}[".format(first_coord["chrom"], first_coord["pos"])
        else:  # First shard is - and second is +
            first_alt = "[{}:{}[N".format(second_coord["chrom"], second_coord["pos"])
            second_alt = "[{}:{}[N".format(first_coord["chrom"], first_coord["pos"])
    return first_alt, second_alt


def getStrand(breakend, is_first=None):
    """
    Return strand from breakend record.

    :param breakend: Breakend defining a part of an RNA fusion.
    :type breakend: anacore.vcf.VCFRecord
    :param is_first: The shard is the first in RNA.
    :type is_first: boolean
    :return: Strand from breakend record.
    :rtype: str
    """
    if is_first is None:
        is_first = True if "RNA_FIRST" in breakend.info else False
    strand = None
    if breakend.alt[0].startswith("[") or breakend.alt[0].startswith("]"):
        strand = "-" if is_first else "+"
    else:
        strand = "+" if is_first else "-"
    return strand


def getCoordStr(breakend, is_first=None):
    """
    Return coordinates elements (chrom, pos and strand) a from breakend record.

    :param breakend: Breakend defining a part of an RNA fusion.
    :type breakend: anacore.vcf.VCFRecord
    :param is_first: The shard is the first in RNA.
    :type is_first: boolean
    :return: Coordinates in dict format {"chrom": "CHROM", "pos": "POS", "strand":"STRAND"}.
    :rtype: dict
    """
    return {"chrom": breakend.chrom, "pos": breakend.pos, "strand": getStrand(breakend, is_first)}


class FusionFileReader(object):
    """Factory to identify and return the dedicated to the software used to produce the fusions file."""

    @staticmethod
    def factory(filepath, *args, **kwargs):
        """
        Return the parser dedicated to the software used to produce the fusions file.

        :param filepath: Path to the file (format: TSV).
        :type filepath: str
        :param args: Additional arguments.
        :type args: list
        :param kwargs: Additional keyword arguments.
        :type kwargs: dict
        :return: The parser dedicated to the software used to produce the fusions file.
        :rtype: ArribaIO or FusionCatcherIO or STARFusionIO
        """
        if STARFusionIO.isValid(filepath):
            return STARFusionIO(filepath, *args, **kwargs)
        elif FusionCatcherIO.isValid(filepath):
            return FusionCatcherIO(filepath, *args, **kwargs)
        elif ArribaIO.isValid(filepath):
            return ArribaIO(filepath, *args, **kwargs)
        elif BreakendVCFIO.isValid(filepath):
            return BreakendVCFIO(filepath, *args, **kwargs)
        else:
            raise IOError("The file {} does not have a valid format for 'FusionFileReader'.".format(filepath))


class FusionCatcherIO(HashedSVIO):
    """Class to read and write fusions in fusionCatcher TSV output format (see https://github.com/ndaniel/fusioncatcher/blob/master/doc/manual.md#output-data). The records are in anacore.vcf.VCFRecord format and a fusion is represented by 2 records (breakend for the 5' shard and breakend for the 3' shard)."""

    titles = [
        "Gene_1_symbol(5end_fusion_partner)",
        "Gene_2_symbol(3end_fusion_partner)",
        "Fusion_description",
        "Counts_of_common_mapping_reads",
        "Spanning_pairs",
        "Spanning_unique_reads",
        "Longest_anchor_found",
        "Fusion_finding_method",
        "Fusion_point_for_gene_1(5end_fusion_partner)",
        "Fusion_point_for_gene_2(3end_fusion_partner)",
        "Gene_1_id(5end_fusion_partner)",
        "Gene_2_id(3end_fusion_partner)",
        "Exon_1_id(5end_fusion_partner)",
        "Exon_2_id(3end_fusion_partner)",
        "Fusion_sequence",
        "Predicted_effect"
    ]

    def __init__(self, filepath, mode="r", annot_field="FCANN", sample_name="sample"):
        """
        Build and return an instance of FusionCatcherIO.

        :param filepath: The filepath.
        :type filepath: str
        :param mode: Mode to open the file ('r', 'w', 'a').
        :type mode: str
        :param annot_field: The tag for the field used to store annotations.
        :type annot_field: str
        :param sample_name: The name of the sample.
        :type sample_name: str
        :return: The new instance.
        :rtype: FusionCatcherIO
        """
        super().__init__(filepath, mode, "\t")
        self.annot_field = annot_field
        self.sample_name = sample_name
        if self.titles is None:
            self.titles = FusionCatcherIO.titles

    def _parseLine(self):
        """
        Return a structured record from the current line.

        :return: The record defined by the current line.
        :rtype: dict
        """
        fusion_record = super()._parseLine()
        return self._dictToBNDRecords(fusion_record)

    def _dictToBNDRecords(self, fusion_record):
        """
        Return the two breakends records from the fusion record.

        :param fusion_record: The fusion record.
        :type fusion_record: dict
        :return: The breakend record for the 5' shard of fusion and the breakend record for the 3' shard.
        :rtype: (anacore.vcf.VCFRecord, anacore.vcf.VCFRecord)
        """
        filters_field = [] if fusion_record["Fusion_description"] == "" else sorted([elt.strip() for elt in fusion_record["Fusion_description"].split(",")])
        if fusion_record["Spanning_unique_reads"] == "0":
            filters_field.append("Imprecise")
            filters_field.sort()
        sample_field = {
            self.sample_name: {
                "PR": int(fusion_record["Spanning_pairs"]),
                "SR": int(fusion_record["Spanning_unique_reads"]),
                "CM": int(fusion_record["Counts_of_common_mapping_reads"]),
                "LA": int(fusion_record["Longest_anchor_found"])
            }
        }
        format_field = sorted(sample_field[self.sample_name].keys())
        info_sources = None if fusion_record["Fusion_finding_method"] == "" else [elt.strip() for elt in fusion_record["Fusion_finding_method"].split(";")]
        first_id = str(uuid.uuid4())
        first_coord = getCoordDictFromCoordStr(fusion_record["Fusion_point_for_gene_1(5end_fusion_partner)"])
        second_id = str(uuid.uuid4())
        second_coord = getCoordDictFromCoordStr(fusion_record["Fusion_point_for_gene_2(3end_fusion_partner)"])
        first_alt, second_alt = getAltFromCoord(first_coord, second_coord)
        # First breakend
        first_info = {
            "MATEID": [second_id],
            "SVTYPE": "BND",
            "SOURCES": info_sources,
            "RNA_FIRST": True,
            self.annot_field: [{
                "SYMBOL": fusion_record["Gene_1_symbol(5end_fusion_partner)"],
                "Gene": fusion_record["Gene_1_id(5end_fusion_partner)"],
                "EXON": fusion_record["Exon_1_id(5end_fusion_partner)"],
                "Effect": fusion_record["Predicted_effect"]
            }]
        }
        first_bnd = VCFRecord(
            first_coord["chrom"],
            first_coord["pos"],
            first_id,
            "N",  # ref
            [first_alt],  # alt
            pFilter=filters_field,
            info=first_info,
            pFormat=format_field,
            samples=sample_field
        )
        # Second breakends
        second_info = {
            "MATEID": [first_id],
            "SVTYPE": "BND",
            "SOURCES": info_sources,
            "RNA_CONTIG": fusion_record["Fusion_sequence"],
            self.annot_field: [{
                "SYMBOL": fusion_record["Gene_2_symbol(3end_fusion_partner)"],
                "Gene": fusion_record["Gene_2_id(3end_fusion_partner)"],
                "EXON": fusion_record["Exon_2_id(3end_fusion_partner)"],
                "Effect": fusion_record["Predicted_effect"]
            }]
        }
        second_bnd = VCFRecord(
            second_coord["chrom"],
            second_coord["pos"],
            second_id,
            "N",  # ref
            [second_alt],  # alt
            pFilter=filters_field,
            info=second_info,
            pFormat=format_field,
            samples=sample_field
        )
        return [first_bnd, second_bnd]

    def write(self, breakends):
        """
        Write record line in file.

        :param record: The first breakend and the second breakend in pair.
        :type record: (anacore.vcf.VCFRecord, anacore.vcf.VCFRecord)
        """
        super().write(self._BNDRecordsToDict(breakends))

    def _BNDRecordsToDict(self, breakends):
        """
        Return the FusionCatcher dict from the two breakends of a fusion.

        :param breakends: The two breakends of a fusion.
        :type breakends: (anacore.vcf.VCFRecord, anacore.vcf.VCFRecord)
        :return: The FusionCatcher dict.
        :rtype: dict
        """
        first, second = breakends
        first_coord = getCoordStr(first, True)
        second_coord = getCoordStr(second, False)
        return {
            "Gene_1_symbol(5end_fusion_partner)": first.info[self.annot_field][0]["SYMBOL"],
            "Gene_2_symbol(3end_fusion_partner)": second.info[self.annot_field][0]["SYMBOL"],
            "Fusion_description": ",".join([elt for elt in first.filter if elt != "Imprecise"]),
            "Counts_of_common_mapping_reads": first.samples[self.sample_name]["CM"],
            "Spanning_pairs": first.samples[self.sample_name]["PR"],
            "Spanning_unique_reads": first.samples[self.sample_name]["SR"],
            "Longest_anchor_found": first.samples[self.sample_name]["LA"],
            "Fusion_finding_method": ";".join(first.info["SOURCES"]),
            "Fusion_point_for_gene_1(5end_fusion_partner)": "{}:{}:{}".format(
                first_coord["chrom"],
                first_coord["pos"],
                first_coord["strand"]
            ),
            "Fusion_point_for_gene_2(3end_fusion_partner)": "{}:{}:{}".format(
                second_coord["chrom"],
                second_coord["pos"],
                second_coord["strand"]
            ),
            "Gene_1_id(5end_fusion_partner)": first.info[self.annot_field][0]["Gene"],
            "Gene_2_id(3end_fusion_partner)": second.info[self.annot_field][0]["Gene"],
            "Exon_1_id(5end_fusion_partner)": first.info[self.annot_field][0]["EXON"],
            "Exon_2_id(3end_fusion_partner)": second.info[self.annot_field][0]["EXON"],
            "Fusion_sequence": second.info["RNA_CONTIG"],
            "Predicted_effect": second.info[self.annot_field][0]["Effect"]
        }

    @staticmethod
    def isValid(filepath):
        """
        Return true if the file is in FusionCatcher output format.

        :param filepath: The file path.
        :type filepath: str
        :return: True if the file is in FusionCatcher output format.
        :rtype: bool
        """
        is_valid = False
        try:
            with SVIO(filepath) as reader:
                mandatory_fields = set(FusionCatcherIO.titles) - {"Predicted_effect"}
                if len(mandatory_fields - set(reader.titles)) == 0:
                    is_valid = True
        except FileNotFoundError:
            raise
        except Exception:
            pass
        return is_valid

    @staticmethod
    def setVCFHeader(vcf_io, annotation_field="FCANN"):
        """
        Set header of an AnnotVCFIO to write FusionCatcher records in VCF output. This function pre-set ANN_titles, info and format used by FusionCatcherIO records.

        :param vcf_io: The VCF used to write FusionCatcher records.
        :type vcf_io: anacore.annotVcf.AnnotVCFIO
        :param annotation_field: INFO field used for store annotations.
        :type annotation_field: str
        """
        # INFO
        vcf_io.info = {
            "MATEID": HeaderInfoAttr("MATEID", type="String", number="A", description="ID of mate breakend."),
            "SVTYPE": HeaderInfoAttr("SVTYPE", type="String", number="1", description="Type of structural variant."),
            "SOURCES": HeaderInfoAttr("SOURCES", type="String", number=".", description="Aligning method used for mapping the reads and finding the fusion genes. Here are two methods used which are: (i) BOWTIE = only Bowtie aligner is used for mapping the reads on the genome and exon-exon fusion junctions, (ii) BOWTIE+BLAT = Bowtie aligner is used for mapping reads on the genome and BLAT is used for mapping reads for finding the fusion junction, (iii) BOWTIE+STAR = Bowtie aligner is used for mapping reads on the genome and STAR is used for mapping reads for finding the fusion junction, (iv) BOWTIE+BOWTIE2 = Bowtie aligner is used for mapping reads on the genome and Bowtie2 is used for mapping reads for finding the fusion junction."),
            "RNA_FIRST": HeaderInfoAttr("RNA_FIRST", type="Flag", number="0", description="For RNA fusions, this break-end is 5' in the fusion transcript."),
            "RNA_CONTIG": HeaderInfoAttr("RNA_CONTIG", type="String", number="1", description="The sequence of the breakend spanning contig."),
            annotation_field: HeaderInfoAttr(annotation_field, type="String", number=".", description="Consequence annotations. Format: SYMBOL|Gene|EXON|Effect.")
        }
        # ANN_titles
        vcf_io.ANN_titles = ["SYMBOL", "Gene", "EXON", "Effect"]
        # FORMAT
        vcf_io.format = {
            "PR": HeaderFormatAttr("PR", type="Integer", number="1", description="Count of pairs of reads supporting the fusion (including also the multimapping reads)."),
            "SR": HeaderFormatAttr("SR", type="Integer", number="1", description="Count of unique reads (i.e. unique mapping positions) mapping on the fusion junction. Shortly, here are counted all the reads which map on fusion junction minus the PCR duplicated reads."),
            "CM": HeaderFormatAttr("CM", type="Integer", number="1", description="Count of reads mapping simultaneously on both genes which form the fusion gene. This is an indication how similar are the DNA/RNA sequences of the genes forming the fusion gene (i.e. what is their homology because highly homologous genes tend to appear show as candidate fusion genes). In case of completely different sequences of the genes involved in forming a fusion gene then here it is expected to have the value zero."),
            "LA": HeaderFormatAttr("LA", type="Integer", number="1", description="Longest anchor (hangover) found among the unique reads mapping on the fusion junction.")
        }


starfusion_filters = {
    "GTEx_recurrent_StarF2019": HeaderFilterAttr(
        "GTEx_recurrent_StarF2019",
        "Fusions found recurrently in GTEx as mined using STAR-Fusion v1.5.0"
    ),
    "BodyMap": HeaderFilterAttr(
        "BodyMap",
        "Fusions found by STAR-Fusion as applied to the Illumina Human Body Map reference data"
    ),
    "DGD_PARALOGS": HeaderFilterAttr(
        "DGD_PARALOGS",
        "Duplicated genes as per the Duplicated Genes Database"
    ),
    "HGNC_GENEFAM": HeaderFilterAttr(
        "HGNC_GENEFAM",
        "HGNC gene family membership as per ftp://ftp.ebi.ac.uk/pub/databases/genenames/genefam_list.txt.gz"
    ),
    "Greger_Normal": HeaderFilterAttr(
        "Greger_Normal",
        "Fusion transcripts (mostly from tandem genes) detected based on analysis of RNA-Seq from 1000 genomes project samples. List derived from Greger et al. PLOS One, 2014"
    ),
    "Babiceanu_Normal": HeaderFilterAttr(
        "Babiceanu_Normal",
        "Recurrent chimeric fusion RNAs in non-cancer tissues and cells as per Babiceanu et al. NAR, 2016"
    ),
    "ConjoinG": HeaderFilterAttr(
        "ConjoinG",
        "Fused transcripts derived from the Conjoined Gene Database"
    ),
    "Imprecise": HeaderFilterAttr(
        "Imprecise",
        "RNA fusion candidates for which no spanning contig was found."
    )
}


class STARFusionIO(HashedSVIO):
    """Class to read and write fusions in STAR-Fusion TSV output format (see https://github.com/STAR-Fusion/STAR-Fusion/wiki#Outputs). The records are in anacore.vcf.VCFRecord format and a fusion is represented by 2 records (breakend for the 5' shard and breakend for the 3' shard)."""

    titles = [
        "FusionName",
        "JunctionReadCount",
        "SpanningFragCount",
        "SpliceType",
        "LeftGene",
        "LeftBreakpoint",
        "RightGene",
        "RightBreakpoint",
        "JunctionReads",
        "SpanningFrags",
        "LargeAnchorSupport",
        "FFPM",
        "LeftBreakDinuc",
        "LeftBreakEntropy",
        "RightBreakDinuc",
        "RightBreakEntropy",
        "annots"
    ]

    def __init__(self, filepath, mode="r", annot_field="FCANN", sample_name="sample"):
        """
        Build and return an instance of STARFusionIO.

        :param filepath: The filepath.
        :type filepath: str
        :param mode: Mode to open the file ('r', 'w', 'a').
        :type mode: str
        :param annot_field: The tag for the field used to store annotations.
        :type annot_field: str
        :param sample_name: The name of the sample.
        :type sample_name: str
        :return: The new instance.
        :rtype: STARFusionIO
        """
        super().__init__(filepath, mode=mode, separator="\t", title_starter="#")
        self.annot_field = annot_field
        self.sample_name = sample_name
        if self.titles is None:
            self.titles = STARFusionIO.titles

    def _parseLine(self):
        """
        Return a structured record from the current line.

        :return: The record defined by the current line.
        :rtype: dict
        """
        fusion_record = super()._parseLine()
        return self._dictToBNDRecords(fusion_record)

    def _dictToBNDRecords(self, fusion_record):
        """
        Return the two breakends records from the fusion record.

        :param fusion_record: The fusion record.
        :type fusion_record: dict
        :return: The breakend record for the 5' shard of fusion and the breakend record for the 3' shard.
        :rtype: (anacore.vcf.VCFRecord, anacore.vcf.VCFRecord)
        """
        breakends = []
        large_anchor_support = "1" if fusion_record["LargeAnchorSupport"] == "YES_LDAS" else "0"
        info_annot_tags = [elt.strip() for elt in json.loads(fusion_record["annots"])]
        filters_field = sorted(list(set(starfusion_filters) & set(info_annot_tags)))
        if fusion_record["JunctionReadCount"] == "0":
            filters_field.append("Imprecise")
            filters_field.sort()
        if len(filters_field) == 0:
            filters_field = ["PASS"]
        coord_list = []
        for side, mate_side in [("Left", "Right"), ("Right", "Left")]:
            id = str(uuid.uuid4())
            coord = getCoordDictFromCoordStr(fusion_record[side + "Breakpoint"])
            coord_list.append(coord)
            # INFO
            gene_symbol, gene_id = fusion_record[side + "Gene"].split("^")
            info = {
                "SVTYPE": "BND",
                "BREAK_DINUC": fusion_record[side + "BreakDinuc"],
                "BREAK_ENTROPY": fusion_record[side + "BreakEntropy"],
                "SPLICE_TYPE": fusion_record["SpliceType"],
                self.annot_field: [{
                    "SYMBOL": gene_symbol,
                    "Gene": gene_id,
                    "Tags": "&".join(info_annot_tags)
                }]
            }
            # SAMPLES
            samples_field = {
                self.sample_name: {
                    "PR": int(fusion_record["SpanningFragCount"]),
                    "SR": int(fusion_record["JunctionReadCount"]),
                    "hasLAS": large_anchor_support,
                    "FFPM": float(fusion_record["FFPM"])
                }
            }
            if side == "Left":
                info["RNA_FIRST"] = True
                if "JunctionReads" in fusion_record:
                    samples_field[self.sample_name]["JRL"] = fusion_record["JunctionReads"]
                if "SpanningFrags" in fusion_record:
                    samples_field[self.sample_name]["SFL"] = fusion_record["SpanningFrags"]
            # Record
            breakends.append(
                VCFRecord(
                    coord["chrom"],
                    coord["pos"],
                    id,
                    "N",  # ref
                    [],  # alt
                    info=info,
                    pFilter=filters_field,
                    pFormat=sorted(samples_field[self.sample_name].keys()),
                    samples=samples_field
                )
            )
        left_alt, right_alt = getAltFromCoord(*coord_list)
        breakends[0].alt = [left_alt]
        breakends[1].alt = [right_alt]
        breakends[0].info["MATEID"] = [breakends[1].id]
        breakends[1].info["MATEID"] = [breakends[0].id]
        return breakends

    def write(self, breakends):
        """
        Write record line in file.

        :param record: The first breakend and the second breakend in pair.
        :type record: (anacore.vcf.VCFRecord, anacore.vcf.VCFRecord)
        """
        super().write(self._BNDRecordsToDict(breakends))

    def _BNDRecordsToDict(self, breakends):
        """
        Return the STAR-fusion dict from the two breakends of a fusion.

        :param breakends: The two breakends of a fusion.
        :type breakends: (anacore.vcf.VCFRecord, anacore.vcf.VCFRecord)
        :return: The STAR-fusion dict.
        :rtype: dict
        """
        first, second = breakends
        first_coord = getCoordStr(first, True)
        second_coord = getCoordStr(second, False)
        return {
            "FusionName": "{}--{}".format(
                first.info[self.annot_field][0]["SYMBOL"],
                second.info[self.annot_field][0]["SYMBOL"],
            ),
            "JunctionReadCount": first.samples[self.sample_name]["SR"],
            "SpanningFragCount": first.samples[self.sample_name]["PR"],
            "SpliceType": first.info["SPLICE_TYPE"],
            "LeftGene": "{}^{}".format(
                first.info[self.annot_field][0]["SYMBOL"],
                first.info[self.annot_field][0]["Gene"]
            ),
            "LeftBreakpoint": "{}:{}:{}".format(
                first_coord["chrom"],
                first_coord["pos"],
                first_coord["strand"]
            ),
            "RightGene": "{}^{}".format(
                second.info[self.annot_field][0]["SYMBOL"],
                second.info[self.annot_field][0]["Gene"]
            ),
            "RightBreakpoint": "{}:{}:{}".format(
                second_coord["chrom"],
                second_coord["pos"],
                second_coord["strand"]
            ),
            "JunctionReads": "" if "JRL" not in first.samples[self.sample_name] else first.samples[self.sample_name]["JRL"],
            "SpanningFrags": "" if "SFL" not in first.samples[self.sample_name] else first.samples[self.sample_name]["SFL"],
            "LargeAnchorSupport": "YES_LDAS" if first.samples[self.sample_name]["hasLAS"] == "1" else "NO_LDAS",
            "FFPM": first.samples[self.sample_name]["FFPM"],
            "LeftBreakDinuc": first.info["BREAK_DINUC"],
            "LeftBreakEntropy": first.info["BREAK_ENTROPY"],
            "RightBreakDinuc": second.info["BREAK_DINUC"],
            "RightBreakEntropy": second.info["BREAK_ENTROPY"],
            "annots": json.dumps(first.info[self.annot_field][0]["Tags"].split("&"))
        }

    @staticmethod
    def isValid(filepath):
        """
        Return true if the file is in STAR-Fusion output format.

        :param filepath: The file path.
        :type filepath: str
        :return: True if the file is in STAR-Fusion output format.
        :rtype: bool
        """
        is_valid = False
        try:
            with SVIO(filepath, title_starter="#") as reader:
                mandatory_fields = set(STARFusionIO.titles) - {"JunctionReads", "SpanningFrags"}
                if len(mandatory_fields - set(reader.titles)) == 0:  # All mandatory fields of Arriba are in reader
                    is_valid = True
        except FileNotFoundError:
            raise
        except Exception:
            pass
        return is_valid

    @staticmethod
    def setVCFHeader(vcf_io, annotation_field="FCANN"):
        """
        Set header of an AnnotVCFIO to write STAR-Fusion records in VCF output. This function pre-set ANN_titles, info and format used by STARFusionIO records.

        :param vcf_io: The VCF used to write STAR-Fusion records.
        :type vcf_io: anacore.annotVcf.AnnotVCFIO
        :param annotation_field: INFO field used for store annotations.
        :type annotation_field: str
        """
        # FILTERS
        vcf_io.filter = starfusion_filters
        # INFO
        vcf_io.info = {
            "MATEID": HeaderInfoAttr("MATEID", type="String", number="A", description="ID of mate breakend."),
            "SVTYPE": HeaderInfoAttr("SVTYPE", type="String", number="1", description="Type of structural variant."),
            "RNA_FIRST": HeaderInfoAttr("RNA_FIRST", type="Flag", number="0", description="For RNA fusions, this breakend is 5' in the fusion transcript."),  # In the fusion transcript, the 'LeftGene' is always 5' relative to the 'RightGene' (https://groups.google.com/forum/#!topic/star-fusion/9rBLT-v5JHI)
            "BREAK_DINUC": HeaderInfoAttr("BREAK_DINUC", type="String", number="1", description="Dinucleotides flanking the breakend of the fragment excluded by the fusion."),
            "BREAK_ENTROPY": HeaderInfoAttr("BREAK_ENTROPY", type="Float", number="1", description="Shannon entropy of the 15 exonic bases flanking the breakpoint. The maximum entropy is 2, representing highest complexity. The lowest would be zero (involving a 15 base mononucleotide run). Low entropy sites should generally be treated as less confident breakpoints."),
            "SPLICE_TYPE": HeaderInfoAttr("SPLICE_TYPE", type="String", number="1", description="Whether the proposed breakpoint occurs at reference exon junctions as provided by the reference transcript structure annotations (ex. gencode)."),
            annotation_field: HeaderInfoAttr(annotation_field, type="String", number=".", description="Annotation generated by FusionAnnotator (see: https://github.com/FusionAnnotator/CTAT_HumanFusionLib/wiki). Format: SYMBOL|Gene|Tags")
        }
        # ANN_titles
        vcf_io.ANN_titles = ["SYMBOL", "Gene", "Tags"]
        # FORMAT
        vcf_io.format = {
            "PR": HeaderFormatAttr("PR", type="Integer", number="1", description="Number of RNA-Seq fragments that encompass the fusion junction such that one read of the pair aligns to a different gene than the other paired-end read of that fragment (SpanningFragCount)."),
            "SR": HeaderFormatAttr("SR", type="Integer", number="1", description="Number of RNA-Seq fragments containing a read that aligns as a split read at the site of the putative fusion junction (JunctionReadCount)."),
            "hasLAS": HeaderFormatAttr("hasLAS", type="String", number="1", description="This column indicates whether there are split reads that provide 'long' (set to length of 25 bases) alignments on both sides of the putative breakpoint (LargeAnchorSupport)."),
            "FFPM": HeaderFormatAttr("FFPM", type="Float", number="1", description="Normalized measures of the fusion-supporting rna-seq fragments (fusion fragments per million total reads)."),
            "JRL": HeaderFormatAttr("JRL", type="String", number="1", description="RNA-Seq fragments containing a read that aligns as a split read at the site of the putative fusion junction."),
            "SFL": HeaderFormatAttr("SFL", type="String", number="1", description="RNA-Seq fragments that encompass the fusion junction such that one read of the pair aligns to a different gene than the other paired-end read of that fragment.")
        }


arriba_event_level_filters = {
    "merge_adjacent": HeaderFilterAttr(
        "merge_adjacent",
        "When multiple alternative alignments with equal quality are possible, STAR does not always choose the same one. This leads to multiple breakpoints in close proximity, all of which arise from one and the same event. This filter merges all breakpoints in a window of 5 bp to the breakpoint with the highest number of supporting reads."
    ),
    "non_coding_neighbors": HeaderFilterAttr(
        "non_coding_neighbors",
        "Genes which are not well studied suffer from incomplete annotation. Many exons are annotated as separate genes even though they might actually be part of one and the same gene. Predicted genes named RP11-... are common examples for this. When poorly understood genes lie next to each other on the same strand, this would frequently lead to false positive predictions of deletions, because the transcripts that span both genes give rise to reads, which resemble focal deletions. This filter discards deletions, which are predicted between two neighboring genes, if both genes are non-coding or one breakpoint is intergenic."
    ),
    "intragenic_exonic": HeaderFilterAttr(
        "intragenic_exonic",
        "Since exons usually make up only a small fraction of a gene, it is more likely that a genomic rearrangement starts and ends in intronic regions. On the transcriptomic level, this manifests as breakpoints at splice-sites or in introns. Many candidates found by STAR have both breakpoints within exons of the same gene. This is particularly true for intragenic events, which are prone to in vitro artifacts. This filter removes intragenic events, if both breakpoints are in exons and more than 80% of the region between the breakpoints is intronic, such that it should be very unlikely that both breakpoints are located inside exons."
    ),
    "min_support": HeaderFilterAttr(
        "min_support",
        "This filter discards all events with few reads."
    ),
    "relative_support": HeaderFilterAttr(
        "relative_support",
        "With increasing expression of a gene the number of chimeric reads mapping to the gene increases, too. Arriba assumes a polynomial relationship between the number of events and the number of supporting reads for a given event. This assumption is based on empirical evidence. An expected value (e-value) is calculated for every event, reflecting how many events with the given number of supporting reads are expected by chance. This filter selects only those events with a low e-value, i.e., that have a high number of supporting reads relative to the overall number of events in a gene. Multiple covariates are taken into account, such as whether a breakpoint is at a splice-site or whether the event is intragenic. The parameters of the polynomial relationship are fixed and were estimated from several hundred RNA-Seq samples of various cancer types."
    ),
    "intronic": HeaderFilterAttr(
        "intronic",
        "This filter discards an event, when none of the supporting reads overlaps an exon (i.e., all alignments are in introns or intergenic regions). True events most often have at least one breakpoint in an exon or at an exon boundary."
    ),
    "known_fusions": HeaderFilterAttr(
        "known_fusions",
        "Events with too few supporting reads but known in cancer."
    ),  # Protection
    "pcr_fusions": HeaderFilterAttr(
        "pcr_fusions",
        "In some tissues certain genes are expressed at very high levels, for example hemoglobin and fibrinogen in blood or collagens in connective tissue. Presumably, the abundance of fragments from such genes increases the chance of unrelated molecules sticking together, which can serve as undesired primers for PCR or may cause the reverse transcriptase enzyme to switch templates. These processes generate a large amount of chimeric fragments in vitro. Such artifactual fusions can be recognized as an extraordinary number of events with breakpoints within exons (rather than at exon boundaries, which is more common for true predictions). This filter eliminates events with genes that are highly expressed (top 0.2%) and have an unbalanced number of split-reads vs. discordant mates or that have an excessive amount of intra-exonic breakpoints."
    ),
    "spliced": HeaderFilterAttr(
        "spliced",
        "This filter recovers events discarded due to a low number of supporting reads, given that both breakpoints of the event are at splice-sites and there is at least one additional event linking the same pair of genes."
    ),  # Protection
    "select_best": HeaderFilterAttr(
        "select_best",
        "If there are multiple breakpoints detected between the same pair of genes, this filter discards all but the most credible one. Events with split reads in both genes are preferred over events with only discordant mates, because in the latter case, the precise breakpoint is unknown. Moreover, events with a higher number of supporting reads are favored."
    ),
    "many_spliced": HeaderFilterAttr(
        "many_spliced",
        "Occassionally, a genomic rearrangement produces multiple alternatively spliced transcripts, all of which are low expressed and therefore discarded by the filters relative_support or min_support. This filter recovers events that were discarded due to too few supporting reads, under the circumstances that there are many events between a pair of genes with at least one of the breakpoints at splice-sites. All events between the pair of genes with at least one spliced breakpoint are recovered."
    ),  # Protection
    "no_genomic_support": HeaderFilterAttr(
        "no_genomic_support",
        "This filter removes events with low confidence, if they are not confirmed by structural variant calls obtained from whole-genome sequencing."
    ),
    "blacklist": HeaderFilterAttr(
        "blacklist",
        "This filter removes events with breakpoint coordinates matching entries in the blacklist. If an event has no split-reads (but only discordant mates), the precise breakpoint coordinates are unknown. In this case the event is discarded, if the breakpoints are within a range of the insert size around the blacklisted coordinates."
    ),
    "short_anchor": HeaderFilterAttr(
        "short_anchor",
        "A chimeric read aligns to some part in one of the fused genes and to some part to the other gene. The anchor of a read is the longer aligned segment. It is more likely to be aligned correctly. True positives often have anchors in both fused genes, whereas alignment artifacts are frequently characterized by only a small segment aligning to one of the genes and all anchors to the other."
    ),
    "end_to_end": HeaderFilterAttr(
        "end_to_end",
        "Theoretically, it should be impossible to observe fusions which only retain the 3' ends of the fusion partners, because the promoter would be missing. In reality, end-to-end fused transcripts are observed, albeit rarely. Most of them are false positives. Therefore, this filter discards such events, unless they have a lot of supporting reads, i.e., split-reads in both fusion partners or split-reads and discordant mates."
    ),
    "no_coverage": HeaderFilterAttr(
        "no_coverage",
        "For intronic and intragenic breakpoints as well as read-through fusions, this filter checks, if there is some coverage in the vicinity of the breakpoint in the normal alignments. Only reads that are not chimeric are considered. When there are no non-chimeric reads near the breakpoint, this is indicative of an alignment artifact."
    ),
    "homologs": HeaderFilterAttr(
        "homologs",
        "This filter discards events between genes that have high sequence homology, which frequently leads to erroneous alignments. Homology is quantified by counting the number of shared 16-mers."
    ),
    "mismappers": HeaderFilterAttr(
        "mismappers",
        "Many alignment artifacts are caused by an excessive number of reference mismatches in close proximity due to sequencing errors or adjacent SNPs. STAR tends to clip reads when it encounters a cluster of mismatches. The clipped segment is then used for a chimeric alignment, which occassionally aligns elsewhere in the genome. The filter mismappers performs a sensitive realignment of both segments. If both segments can be aligned to the same gene (while allowing more mismatches than STAR does), the chimeric alignment is considered to be an artifact. When an important proportion of the supporting reads are classified as being aligned incorrectly, the event is discarded."
    ),
    "genomic_support": HeaderFilterAttr(
        "genomic_support",
        "This filter recovers events which were discarded by previous filters due to few supporting reads, but which can be explained by genomic rearrangements as evidenced by structural variant calls obtained from whole-genome sequencing data. Arriba considers structural variant calls to match with breakpoints seen in transcriptomic data, when the breakpoints are less then 100 kb apart and the orientation of the genomic and transcriptomic breakpoints are identical."
    ),
    "isoforms": HeaderFilterAttr(
        "isoforms",
        "This filter searches for additional isoforms for those gene pairs that are predicted to be fused in proper orientation. Typically there is a major isoform which is expressed at a high level and a few low expressed isoforms."
    ),
    "Imprecise": HeaderFilterAttr(
        "Imprecise",
        "RNA fusion candidates for which no spanning contig was found."
    ),
    "Unstranded": HeaderFilterAttr(
        "Unstranded",
        "RNA fusion is not stranded."
    )
}


class ArribaIO(HashedSVIO):
    """Class to read and write fusions in Arriba TSV output format (see https://github.com/suhrig/arriba/blob/master/documentation/output-files.md). The records are in anacore.vcf.VCFRecord format and a fusion is represented by 2 records (breakend for the 5' shard and breakend for the 3' shard)."""

    titles = [
        "gene1",
        "gene2",
        "strand1(gene/fusion)",
        "strand2(gene/fusion)",
        "breakpoint1",
        "breakpoint2",
        "site1",
        "site2",
        "type",
        "direction1",
        "direction2",
        "split_reads1",
        "split_reads2",
        "discordant_mates",
        "coverage1",
        "coverage2",
        "confidence",
        "closest_genomic_breakpoint1",
        "closest_genomic_breakpoint2",
        "filters",
        "fusion_transcript",
        "reading_frame",
        "peptide_sequence",
        "read_identifiers"
    ]

    def __init__(self, filepath, mode="r", annot_field="FCANN", sample_name="sample"):
        """
        Build and return an instance of ArribaIO.

        :param filepath: The filepath.
        :type filepath: str
        :param mode: Mode to open the file ('r', 'w', 'a').
        :type mode: str
        :param annot_field: The tag for the field used to store annotations.
        :type annot_field: str
        :param sample_name: The name of the sample.
        :type sample_name: str
        :return: The new instance.
        :rtype: ArribaIO
        """
        super().__init__(filepath, mode=mode, separator="\t", title_starter="#")
        self.annot_field = annot_field
        self.sample_name = sample_name
        if self.titles is None:
            self.titles = ArribaIO.titles

    def _parseLine(self):
        """
        Return a structured record from the current line.

        :return: The record defined by the current line.
        :rtype: dict
        """
        fusion_record = super()._parseLine()
        return self._dictToBNDRecords(fusion_record)

    def _dictToBNDRecords(self, fusion_record):
        """
        Return the two breakends records from the fusion record.

        :param fusion_record: The fusion record.
        :type fusion_record: dict
        :return: The breakend record for the 5' shard of fusion and the breakend record for the 3' shard.
        :rtype: (anacore.vcf.VCFRecord, anacore.vcf.VCFRecord)
        """
        is_unstranded = fusion_record["strand1(gene/fusion)"][-1] == "." or fusion_record["strand2(gene/fusion)"][-1] == "."
        event_level_filters = set(arriba_event_level_filters)
        filters_from_file = {elt.strip() for elt in fusion_record["filters"].split(",")}
        filters_field = sorted(list(filters_from_file & event_level_filters))
        read_level_filters = sorted(list(filters_from_file - event_level_filters))
        is_imprecise = False if int(fusion_record["split_reads1"]) + int(fusion_record["split_reads2"]) > 0 else True
        if is_imprecise:
            filters_field.append("Imprecise")
        if is_unstranded:
            filters_field.append("Unstranded")
        if is_imprecise or is_unstranded:
            filters_field.sort()
        breakends = []
        coord_list = []
        for side, mate_side in [("1", "2"), ("2", "1")]:
            id = str(uuid.uuid4())
            strand = fusion_record["strand" + side + "(gene/fusion)"].split("/")[1]
            if is_unstranded:
                strand = "+"
                if side == "1":
                    if fusion_record["direction1"] == "downstream":  # The first in fusion go to the breakpoint: ---->| or |<----
                        strand = "-"
                else:
                    if fusion_record["direction2"] == "upstream":  # The second in fusion go from the breakpoint: <----| or |---->
                        strand = "-"
            curr_breakpoint = fusion_record["breakpoint" + side] + ":" + strand
            coord = getCoordDictFromCoordStr(curr_breakpoint)
            coord_list.append(coord)
            # INFO
            has_frameshift = {"in-frame": "false", "out-of-frame": "true", ".": "."}
            info = {
                "SVTYPE": "BND",
                self.annot_field: [{
                    "SYMBOL": fusion_record["gene" + side],
                    "STRAND": fusion_record["strand" + side + "(gene/fusion)"].split("/")[0],
                    "Site": fusion_record["site" + side],
                    "Type": fusion_record["type"],
                    "GENE_SHARD": fusion_record["direction" + side].replace("stream", ""),
                    "FRAMESHIFT": has_frameshift[fusion_record["reading_frame"]],
                    "Protein_contig": fusion_record["peptide_sequence"].replace("|", "@")
                }]
            }
            if is_imprecise:
                info["IMPRECISE"] = True
            if is_unstranded:
                info["UNSTRANDED"] = True
            if fusion_record["closest_genomic_breakpoint" + side] != ".":
                info["GBP"] = fusion_record["closest_genomic_breakpoint" + side]
            # SAMPLES
            samples_field = {
                self.sample_name: {
                    "PR": int(fusion_record["discordant_mates"]),
                    "SR": int(fusion_record["split_reads1"]) + int(fusion_record["split_reads2"]),
                    "SR1": int(fusion_record["split_reads1"]),
                    "SR2": int(fusion_record["split_reads2"]),
                    "CFD": fusion_record["confidence"],
                    "RFIL": read_level_filters
                }
            }
            if fusion_record["coverage" + side] != ".":
                samples_field[self.sample_name]["DPS"] = int(fusion_record["coverage" + side])
            if fusion_record["read_identifiers"] != ".":
                samples_field[self.sample_name]["SRL"] = fusion_record["read_identifiers"].split(",")
            if side == "1":
                info["RNA_FIRST"] = True
            else:
                if fusion_record["fusion_transcript"] != ".":
                    info["RNA_CONTIG"] = fusion_record["fusion_transcript"].replace("|", "@")
            # Record
            breakends.append(
                VCFRecord(
                    coord["chrom"],
                    coord["pos"],
                    id,
                    "N",  # ref
                    [],  # alt
                    pFilter=["PASS"] if len(filters_field) == 0 else filters_field,
                    info=info,
                    pFormat=sorted(samples_field[self.sample_name].keys()),
                    samples=samples_field
                )
            )
        left_alt, right_alt = getAltFromCoord(*coord_list)
        breakends[0].alt = [left_alt]
        breakends[1].alt = [right_alt]
        breakends[0].info["MATEID"] = [breakends[1].id]
        breakends[1].info["MATEID"] = [breakends[0].id]
        return breakends

    def write(self, breakends):
        """
        Write record line in file.

        :param record: The first breakend and the second breakend in pair.
        :type record: (anacore.vcf.VCFRecord, anacore.vcf.VCFRecord)
        """
        super().write(self._BNDRecordsToDict(breakends))

    def _BNDRecordsToDict(self, breakends):
        """
        Return the Arriba dict from the two breakends of a fusion.

        :param breakends: The two breakends of a fusion.
        :type breakends: (anacore.vcf.VCFRecord, anacore.vcf.VCFRecord)
        :return: The Arriba dict.
        :rtype: dict
        """
        first, second = breakends
        first_coord = getCoordStr(first, True)
        second_coord = getCoordStr(second, False)
        fs_to_reading_frame = {".": ".", "false": "in-frame", "true": "out-of-frame"}
        arriba_filters = ""
        if first.filter:
            excluded = {"Imprecise", "Unstranded"}
            arriba_filters = ",".join(sorted(set(first.filter) - excluded))
        if "RFIL" in first.samples[self.sample_name] and first.samples[self.sample_name]["RFIL"]:
            arriba_filters += ",".join(first.samples[self.sample_name]["RFIL"])
        return {
            "gene1": first.info[self.annot_field][0]["SYMBOL"],
            "gene2": second.info[self.annot_field][0]["SYMBOL"],
            "strand1(gene/fusion)": "{}/{}".format(
                first.info[self.annot_field][0]["STRAND"],
                "." if "UNSTRANDED" in first.info else first_coord["strand"]
            ),
            "strand2(gene/fusion)": "{}/{}".format(
                second.info[self.annot_field][0]["STRAND"],
                "." if "UNSTRANDED" in first.info else second_coord["strand"]
            ),
            "breakpoint1": "{}:{}".format(first_coord["chrom"], first_coord["chrom"]),
            "breakpoint2": "{}:{}".format(second_coord["chrom"], second_coord["chrom"]),
            "site1": first.info[self.annot_field][0]["Site"],
            "site2": second.info[self.annot_field][0]["Site"],
            "type": second.info[self.annot_field][0]["Type"],
            "direction1": first.info[self.annot_field][0]["GENE_SHARD"] + "stream",
            "direction2": second.info[self.annot_field][0]["GENE_SHARD"] + "stream",
            "split_reads1": first.samples[self.sample_name]["SR1"],
            "split_reads2": first.samples[self.sample_name]["SR2"],
            "discordant_mates": first.samples[self.sample_name]["PR"],
            "coverage1": first.samples[self.sample_name]["DPS"],
            "coverage2": second.samples[self.sample_name]["DPS"],
            "confidence": first.samples[self.sample_name]["CFD"],
            "closest_genomic_breakpoint1": "." if "GBP" not in first.info else first.info["GBP"],
            "closest_genomic_breakpoint2": "." if "GBP" not in second.info else second.info["GBP"],
            "filters": arriba_filters,
            "fusion_transcript": "." if "RNA_CONTIG" not in first.info else first.info["RNA_CONTIG"].replace("@", "|"),
            "reading_frame": fs_to_reading_frame[second.info[self.annot_field][0]["FRAMESHIFT"]],
            "peptide_sequence": "." if first.info[self.annot_field][0]["Protein_contig"] is None else first.info[self.annot_field][0]["Protein_contig"].replace("@", "|"),
            "read_identifiers": "." if "SRL" not in first.samples[self.sample_name] or len(first.samples[self.sample_name]["SRL"]) == 0 else first.samples[self.sample_name]["SRL"],
        }

    @staticmethod
    def isValid(filepath):
        """
        Return true if the file is in Arriba output format.

        :param filepath: The file path.
        :type filepath: str
        :return: True if the file is in Arriba output format.
        :rtype: bool
        """
        is_valid = False
        try:
            with SVIO(filepath, title_starter="#") as reader:
                if len(set(ArribaIO.titles) - set(reader.titles)) == 0:  # All mandatory fields of Arriba are in reader
                    is_valid = True
        except FileNotFoundError:
            raise
        except Exception:
            pass
        return is_valid

    @staticmethod
    def setVCFHeader(vcf_io, annotation_field="FCANN"):
        """
        Set header of an AnnotVCFIO to write Arriba records in VCF output. This function pre-set ANN_titles, info and format used by ArribaIO records.

        :param vcf_io: The VCF used to write Arriba records.
        :type vcf_io: anacore.annotVcf.AnnotVCFIO
        :param annotation_field: INFO field used for store annotations.
        :type annotation_field: str
        """
        # INFO
        vcf_io.info = {
            "GBP": HeaderInfoAttr("GBP", type="String", number="1", description="The coordinates of the genomic breakpoint which is closest to the transcriptomic breakpoint."),
            "IMPRECISE": HeaderInfoAttr("IMPRECISE", type="Flag", number="0", description="Imprecise structural variation."),
            "MATEID": HeaderInfoAttr("MATEID", type="String", number="A", description="ID of mate breakend."),
            "RNA_CONTIG": HeaderInfoAttr("RNA_CONTIG", type="String", number="1", description="The transcript sequence assembled from the supporting reads of the most highly expressed transcript."),
            "RNA_FIRST": HeaderInfoAttr("RNA_FIRST", type="Flag", number="0", description="For RNA fusions, this break-end is 5' in the fusion transcript."),  # In the fusion transcript, the 'LeftGene' is always 5' relative to the 'RightGene' (https://groups.google.com/forum/#!topic/star-fusion/9rBLT-v5JHI)
            "SVTYPE": HeaderInfoAttr("SVTYPE", type="String", number="1", description="Type of structural variant."),
            "UNSTRANDED": HeaderInfoAttr("UNSTRANDED", type="Flag", number="0", description="Unstranded structural variation."),
            annotation_field: HeaderInfoAttr(annotation_field, type="String", number=".", description="Consequence annotations. Format: SYMBOL|STRAND|Site|Type|GENE_SHARD|FRAMESHIFT|Protein_contig")
        }
        # ANN_titles
        vcf_io.ANN_titles = ["SYMBOL", "STRAND", "Site", "Type", "GENE_SHARD", "FRAMESHIFT", "Protein_contig"]
        # FILTERS
        vcf_io.filter = arriba_event_level_filters
        # FORMAT
        vcf_io.format = {
            "PR": HeaderFormatAttr("PR", type="Integer", number="1", description="Number of RNA-Seq fragments that encompass the fusion junction such that one read of the pair aligns to a different gene than the other paired-end read of that fragment."),
            "SR": HeaderFormatAttr("SR", type="Integer", number="1", description="Number of RNA-Seq fragments containing a read that aligns as a split read at the site of the putative fusion junction."),
            "SR1": HeaderFormatAttr("SR1", type="Integer", number="1", description="Number of RNA-Seq fragments containing a read that aligns as a split read at the site of the putative fusion junction with an anchor on first shard."),
            "SR2": HeaderFormatAttr("SR2", type="Integer", number="1", description="Number of RNA-Seq fragments containing a read that aligns as a split read at the site of the putative fusion junction with an anchor on second shard."),
            "CFD": HeaderFormatAttr("CFD", type="String", number="1", description="Each prediction is assigned one of the confidences low, medium, or high. Several characteristics are taken into account, including: the number of supporting reads, the balance of split reads and discordant mates, the distance between the breakpoints, the type of event, whether the breakpoints are intragenic or not, and whether there are other events which corroborate the prediction, e.g. multiple isoforms or balanced translocations."),
            "DPS": HeaderFormatAttr("DPS", type="Integer", number="1", description="Coverage near breakpoint. The coverage is calculated as the number of fragments near the breakpoint on the side of the breakpoint that is retained in the fusion transcript. Note that the coverage calculation counts all fragments (even duplicates)."),
            "RFIL": HeaderFormatAttr("RFIL", type="String", number=".", description="Filters which removed one or more of the supporting reads. The number of filtered reads is given in parantheses after the name of the filter. If a filter discarded the event as a whole (all reads), the number of filtered reads is missing."),
            "SRL": HeaderFormatAttr("SRL", type="String", number=".", description="The names of the supporting reads.")
        }


class BreakendVCFIO(AnnotVCFIO):
    """Read and write VCF file containing breakends. Each iteration return a couple of breakends (the first and the second in fusion)."""

    def __init__(self, filepath, mode="r", annot_field=None):
        """
        Return instance of BreakendVCFIO.

        :param filepath: The filepath.
        :type filepath: str
        :param mode: Mode to open the file ('r', 'w', 'a', 'i'). The mode 'i' allow to open file in indexed mode to fetch by region (tabix).
        :type mode: str
        :return: The new instance.
        :rtype: anacore.fusionVcf.BreakendVCFIO
        """
        self.sv_rec_by_id = {}  # By record ID the byte position of start and end of line
        self.pick_reader = None
        if mode in {"r", "i"}:
            if isGzip(filepath):
                self.pick_reader = gzip.open(filepath, "rt")
            else:
                self.pick_reader = open(filepath, "r")
            self.loadIndex()
        self._iter_already_processed = set()
        super().__init__(filepath, mode, annot_field)

    def loadIndex(self):
        """Parse file and store in index the byte positions (start and end) of each record by record ID."""
        self.sv_rec_by_id = {}
        next_start = 0
        line_nb = 1
        line = self.pick_reader.readline()
        while line:
            start = next_start
            next_start = self.pick_reader.tell()
            if not line.startswith("#"):
                end = next_start - 1
                id = line.split(None, 3)[2].strip()
                self.sv_rec_by_id[id] = (start, end, line_nb)
            line = self.pick_reader.readline()
            line_nb += 1

    def close(self):
        """Close file handle."""
        super().close()
        if self.pick_reader is not None:
            self.pick_reader.close()

    def get(self, id):
        """
        Return record with provided ID.

        :param id: record ID.
        :type id: str
        :return: The record.
        :rtype: anacore.vcf.VCFRecord
        """
        # Get current pos
        current = {
            "line": self.current_line,
            "line_nb": self.current_line_nb,
        }
        # Get new record
        start, end, line_nb = self.sv_rec_by_id[id]
        self.pick_reader.seek(start)
        self.current_line = self.pick_reader.read(end - start + 1)
        self.current_line_nb = line_nb
        record = self._parseLine()
        # Jump to previous record
        self.current_line = current["line"]
        self.current_line_nb = current["line_nb"]
        # Return
        return record

    def getSub(self, chr, start, end):
        """
        Return generator on records overlapping the specified region.

        .. warning::
            This method can only be used with an instance of anacore.VCFIO opened
            with mode "i".

        :param chr: Chromosome name of the selected region.
        :type chr: str
        :param start: Start of the selected region (1-based).
        :type start: int
        :param end: End of the selected region (1-based).
        :type end: int
        :return: Fusions (record and his mates) overlappind the specified region.
        :rtype: generator for (anacore.vcf.VCFRecord, [anacore.vcf.VCFRecord, ...])
        """
        for record in super().getSub(chr, start, end):
            mates = []
            for mate_id in record.info["MATEID"]:
                mates.append(self.get(mate_id))
            yield (record, mates)

    def isRecordLine(self, line):
        """
        Return True if the line corresponds to a new record (it is not a comment, an header line or an already processed fusion).

        :param line: The evaluated line.
        :type line: str.
        :return: True if the line corresponds to a record.
        :rtype: bool
        """
        is_record = super().isRecordLine(line)
        if is_record:
            id = line.split("\t", 3)[2].strip()
            mate_info = line.split("\t", 8)[7].strip()
            match = re.search("MATEID\=([^;]+);?", mate_info)
            if match:
                mate_id = decodeInfoValue(match.groups()[0])
                fusion_id = " @@ ".join(sorted([id, mate_id]))
                if fusion_id in self._iter_already_processed:  # Partial pre-filtering for records without multi-alternatives
                    is_record = False
            else:
                is_record = False
        return is_record

    def __iter__(self):
        for record in super().__iter__():
            for mate_idx, mate_id in enumerate(record.info["MATEID"]):
                mate = self.get(mate_id)
                fusion_id = " @@ ".join(sorted([record.id, mate.id]))
                if fusion_id not in self._iter_already_processed:
                    self._iter_already_processed.add(fusion_id)
                    # Change ID
                    record_new_id = record.id
                    alt_record = record
                    if len(record.alt) > 1:
                        record_new_id += "_" + str(mate_idx)  # Record must be splitted for each mate
                        alt_record = getAlleleRecord(self, record, mate_idx)
                        alt_record.info["MATEID"] = [record.info["MATEID"][mate_idx]]
                    mate_new_id = mate.id
                    alt_mate = mate
                    if len(mate.alt) > 1:
                        record_idx = mate.info["MATEID"].index(record.id)
                        mate_new_id += "_" + str(record_idx)  # Record must be splitted for each mate
                        alt_mate = getAlleleRecord(self, mate, record_idx)
                        alt_mate.info["MATEID"] = [mate.info["MATEID"][record_idx]]
                    alt_record.id = record_new_id
                    alt_mate.info["MATEID"] = [record_new_id]
                    alt_mate.id = mate_new_id
                    alt_record.info["MATEID"] = [mate_new_id]
                    # Order by transcript
                    first = alt_record
                    second = alt_mate
                    if "RNA_FIRST" in second.info:
                        first = alt_mate
                        second = alt_record
                    # Return
                    yield first, second

    @staticmethod
    def isValid(filepath):
        """
        Return true if the file is in VCF output format.

        :param filepath: The file path.
        :type filepath: str
        :return: True if the file is in VCF output format.
        :rtype: bool
        """
        is_valid = False
        try:
            with VCFIO(filepath) as reader:
                if "SVTYPE" in reader.info and "MATEID" in reader.info:
                    is_valid = True
        except FileNotFoundError:
            raise
        except Exception:
            pass
        return is_valid

    def write(self, first, second):
        """
        Write fusion record in VCF.

        :param first: The fisrt breakend.
        :type first: anacore.vcf.VCFRecord
        :param second: The second breakend.
        :type second: anacore.vcf.VCFRecord
        """
        super().write(first)
        super().write(second)



# -*- coding: utf-8 -*-
"""Classes and functions for manipulating/processing Tree."""

__author__ = 'Frederic Escudie'
__copyright__ = 'Copyright (C) 2015 INRA'
__license__ = 'GNU General Public License'
__version__ = '1.0.3'
__email__ = 'escudie.frederic@iuct-oncopole.fr'
__status__ = 'prod'

import json


class Node:
    """A node is an element of a tree structure. It is linked with its parent and its children and it is described by several metadata."""

    def __init__(self, name=None, parent_node=None, children_nodes=None, metadata=None):
        """
        Build and return an instance of Node.

        :param name: The node name.
        :type name: str
        :param parent_node: The parent node.
        :type parent_node: Node
        :param children_nodes: List of children nodes.
        :type children_nodes: list
        :param metadata: The metadata.
        :type metadata: dict
        :return: The new instance.
        :rtype: node.Node
        """
        self.children = list()
        if children_nodes is not None:
            for current_child in children_nodes:
                self.addChild(current_child)
        self.name = name
        self.metadata = metadata if metadata is not None else dict()
        self.parent = None
        if parent_node is not None:
            parent_node.addChild(self)

    def __str__(self):
        """
        Return the “informal” or nicely printable string representation of the object.

        :return: The printable string representation of the object.
        :rtype: str
        """
        node_str = "-" if self.name is None else self.name
        node_str += "\n\tParent=" + ("-" if self.parent is None or self.parent.name is None else self.parent.name)
        node_str += "\n\tChilds=" + ", ".join(self.children.keys())
        metadata = list()
        for key in sorted(self.metadata.keys()):
            metadata.append(key + ":" + str(self.metadata[key]))
        node_str += "\n\tMetadata=" + ", ".join(metadata)
        return node_str

    def hasChild(self, name=None):
        """
        Return True if the node has the specified child (name is set) or at least one child (name is None).

        :param name: The name of the searched child.
        :type name: str
        :return: True if the node has the specified child (name is set) or at least one child (name is None).
        :rtype: bool
        """
        if name is None:
            return len(self.children) > 0
        else:
            children_names = [child.name for child in self.children]
            return name in children_names

    def getChildByName(self, name):
        """
        Return the specified child node.

        :param name: Name of the searched child.
        :type name: str
        :return: The child.
        :rtype: node.Node
        """
        selected_child = None
        for child in self.children:
            if child.name is not None and child.name == name:
                selected_child = child
        if selected_child is None:
            raise Exception(
                "The node doesn't have a child named '{}'.".format(name)
            )
        return selected_child

    def getAncestors(self):
        """
        Return the ancestors of the node.

        :return: List ancestors nodes. The nodes are in parent to child order.
        :rtype: list
        """
        ancestors = list()
        if self.parent is not None:
            ancestors.extend(self.parent.getAncestors())
            ancestors.extend([self.parent])
        return ancestors

    def getDescendants(self, depth=1):
        """
        Return the node descendants with the provided depth from the node. Example: depth=1 returns all the children of the node ; depth=2 returns all the grandchildren of the node.

        :param depth: The selected depth.
        :type depth: int
        :return: Descendants.
        :rtype: list
        """
        descendants = list()
        if depth == 1:
            descendants = self.children
        elif depth > 1:
            for child in self.self.children:
                descendants.extend(child.getDescendants(depth - 1))
        return descendants

    def getLeaves(self):
        """
        Return leaves.

        :return: Leaves.
        :rtype: list
        """
        leaves = list()
        if not self.hasChild():
            leaves = [self]
        else:
            for child in self.children:
                leaves.extend(child.getLeaves())
        return leaves

    def getDepth(self):
        """
        Return the depth of the node (= the branch length / = the number of ancestors nodes).

        :return: The depth of the node. The depth for the root element is 0.
        :rtype: int
        """
        if self.parent is None:
            return 0
        else:
            return(self.parent.getDepth() + 1)

    def addChild(self, child):
        """
        Add a node as child.

        :param child: The added node.
        :type child: node.Node
        """
        if child.name is not None:  # Check duplication for named nodes
            children_names = [child.name for child in self.children]
            if child.name in children_names:
                raise Exception(
                    "Duplicated child name '{}' in node '{}'.".format(
                        child.name, self.name
                    )
                )
        child.parent = self
        self.children.append(child)

    @staticmethod
    def fromDict(tree):
        """
        Return nodes representing the tree.

        :param tree: [dict] Each node is dict with following form: {"name": NAME, "metadata":{...}, "children": OTHER_NODE}. Each attribute is optional.
        :return: The node representing the root.
        :rtype: node.Node
        """
        # Current node
        curr_node = Node()
        if "name" in tree:
            curr_node.name = tree["name"]
        if "metadata" in tree:
            for met_key, met_val in tree["metadata"].items():
                curr_node.metadata[met_key] = met_val
        # Chilren
        if "children" in tree and len(tree["children"]) > 0:
            for child in tree["children"]:
                curr_node.addChild(
                    Node.fromDict(child)
                )
        return curr_node

    @staticmethod
    def fromClusterNode(tree, id_to_name=None, distance_tag="dist", _parent_dist_from_root=0, _root_dist_from_leaves=None):
        """
        Return nodes representing the tree.

        :param tree: The tree returned by numpy.
        :type tree: scipy.cluster.hierarchy.ClusterNode
        :param id_to_name: The link between numpy's node id and the node name.
        :type id_to_name: dict
        :param distance_tag: Tag used to store distance in metadata.
        :type distance_tag: str
        :param _parent_dist_from_root: Distance between root and parent node. By default the first node provided is view as root.
        :type _parent_dist_from_root: float
        :param _root_dist_from_leaves: Maximum distance between root and leaves. By default the first node provided is view as root.
        :type _root_dist_from_leaves: float
        :return: The node representing the root.
        :rtype: node.Node
        """
        if _root_dist_from_leaves is None:  # First node is root
            _root_dist_from_leaves = tree.dist
        # Current node
        node_name = None
        if id_to_name is not None and tree.get_id() in id_to_name:
            node_name = id_to_name[tree.get_id()]
        curr_dist_from_root = _root_dist_from_leaves - tree.dist
        curr_node = Node(
            node_name,
            metadata={
                distance_tag: curr_dist_from_root - _parent_dist_from_root,
                "dist_from_root": curr_dist_from_root
            }
        )
        # Chilren
        if tree.get_left():
            curr_node.addChild(
                Node.fromClusterNode(tree.get_left(), id_to_name, distance_tag, curr_dist_from_root, _root_dist_from_leaves)
            )
        if tree.get_right():
            curr_node.addChild(
                Node.fromClusterNode(tree.get_right(), id_to_name, distance_tag, curr_dist_from_root, _root_dist_from_leaves)
            )
        return curr_node

    def toDict(self):
        """
        Return a dictionary representing the instance.

        :return: The dictionary representing the instance with following form: {"name": NAME, "metadata":{...}, "children": OTHER_NODE}.
        :rtype: dict
        """
        return {
            "name": self.name,
            "metadata": self.metadata,
            "children": [child.toDict() for child in self.children]
        }

    def toNewick(self, distance_tag=None):
        """
        Return the representation of the tree rooted by the node in newick format.

        :param distance_tag: The metadata tag for the node distance (default: 'dist'). The distance is not necessary to use this method.
        :type distance_tag: str
        :returns: Newick representation of the tree.
        :rtype: str
        """
        if distance_tag is None:
            distance_tag = "dist"
        node_newick = ""
        if self.hasChild():
            children_newick = list()
            for child in self.children:
                children_newick.append(child.toNewick())
            node_newick = '({})'.format(','.join(children_newick))
        if self.name is not None:
            node_newick += '"' + self.name + '"'
        if distance_tag in self.metadata:
            node_newick += ':{}'.format(self.metadata[distance_tag])
        return node_newick

    def toExtendedNewick(self):
        """
        Return the representation of the tree rooted by the node in extended newick format. In extended newick the distance tag is replaced by a json represetation of the metadata.

        :return: Extended newick representation of the tree.
        :rtype: str
        """
        node_newick = ""
        if self.hasChild():
            children_newick = list()
            for child in self.children:
                children_newick.append(child.toExtendedNewick())
            node_newick = '({})'.format(','.join(children_newick))
        if self.name is not None:
            node_newick += '"' + self.name + '"'
        if len(self.metadata) != 0:
            node_newick += ':{}'.format(json.dumps(self.metadata))
        return node_newick



#!/usr/bin/env python3

__author__ = 'Frederic Escudie'
__copyright__ = 'Copyright (C) 2018 IUCT-O'
__license__ = 'GNU General Public License'
__version__ = '1.6.0'
__email__ = 'escudie.frederic@iuct-oncopole.fr'
__status__ = 'prod'

import os
import sys
import tempfile
import unittest
import uuid

TEST_DIR = os.path.dirname(os.path.abspath(__file__))
PACKAGE_DIR = os.path.dirname(TEST_DIR)
sys.path.append(PACKAGE_DIR)

from anacore.genomicRegion import Gene, Transcript, Exon, Intron, Protein, CDS
from anacore.region import Region
from anacore.sequenceIO import IdxFastaIO


########################################################################
#
# FUNCTIONS
#
########################################################################
class TestTranscript(unittest.TestCase):
    def setUp(self):
        # Forward
        self.fwd = {
            "exon_1": Exon(10, 30, "+", "chr1"),
            "exon_2": Exon(40, 70, "+", "chr1"),
            "exon_3": Exon(80, 100, "+", "chr1"),
            "intron_1": Intron(31, 39, "+", "chr1"),
            "intron_2": Intron(71, 79, "+", "chr1")
        }
        self.fwd["transcript"] = Transcript(
            10, 100, "+", "chr1", children=[
                self.fwd["exon_1"], self.fwd["exon_3"], self.fwd["exon_2"]
            ]
        )
        # Reverse
        self.rvs = {
            "exon_1": Exon(10, 30, "-", "chr1"),
            "exon_2": Exon(40, 70, "-", "chr1"),
            "exon_3": Exon(80, 100, "-", "chr1"),
            "intron_1": Intron(31, 39, "-", "chr1"),
            "intron_2": Intron(71, 79, "-", "chr1")
        }
        self.rvs["transcript"] = Transcript(
            reference="chr1",
            children=[
                self.rvs["exon_1"], self.rvs["exon_3"], self.rvs["exon_2"]
            ]
        )

    def addProtein(self):
        protein_1 = Protein(10, 30, "+", "chr1", "p1")
        protein_2 = Protein(32, 50, "+", "chr1", "p2")
        transcript_1 = Transcript(name="tr1")
        # Empty
        self.assertEqual(
            [elt.name for elt in transcript_1.proteins],
            []
        )
        # Add protein_1
        transcript_1.addProtein(protein_1)
        self.assertEqual(
            [elt.name for elt in transcript_1.proteins],
            ["p1"]
        )  # Check from tr
        self.assertEqual(
            protein_1.transcript,
            transcript_1
        )  # Check from prot
        # Add protein_2
        transcript_1.addProtein(protein_2)
        self.assertEqual(
            [elt.name for elt in transcript_1.proteins],
            ["p1", "p2"]
        )  # Check from tr
        self.assertEqual(
            protein_1.transcript,
            transcript_1
        )  # Check from prot
        self.assertEqual(
            protein_2.transcript,
            transcript_1
        )  # Check from prot

    def delProtein(self):
        protein_1 = Protein(10, 30, "+", "chr1", "p1")
        protein_2 = Protein(32, 50, "+", "chr1", "p2")
        transcript_1 = Transcript(name="tr1", proteins=[protein_1, protein_2])
        # Init
        self.assertEqual(
            [elt.name for elt in transcript_1.proteins],
            ["p1", "p2"]
        )  # Check from tr
        self.assertEqual(
            protein_1.transcript,
            transcript_1
        )  # Check from prot
        self.assertEqual(
            protein_2.transcript,
            transcript_1
        )  # Check from prot
        # Delete protein 2
        transcript_1.delProtein(protein_2)
        self.assertEqual(
            [elt.name for elt in transcript_1.proteins],
            ["p1"]
        )  # Check from tr
        self.assertEqual(
            protein_1.transcript,
            transcript_1
        )  # Check from prot
        self.assertEqual(
            protein_2.transcript,
            None
        )  # Check from prot
        # Delete protein 1
        transcript_1.addProtein(protein_1)
        self.assertEqual(
            [elt.name for elt in transcript_1.proteins],
            []
        )  # Check from tr
        self.assertEqual(
            protein_1.transcript,
            None
        )  # Check from prot
        self.assertEqual(
            protein_2.transcript,
            None
        )  # Check from prot

    def testSetProteins(self):
        # By init
        protein_1 = Protein(10, 30, "+", "chr1", "p1")
        protein_2 = Protein(32, 50, "+", "chr1", "p2")
        transcript_1 = Transcript(name="tr1", proteins=[protein_1, protein_2])
        self.assertEqual(
            [prot.name for prot in transcript_1.proteins],
            [protein_1.name, protein_2.name]
        )
        self.assertEqual(
            [protein_1.transcript.name, protein_2.transcript.name],
            [transcript_1.name, transcript_1.name]
        )
        # By method
        protein_1 = Protein(10, 30, "+", "chr1", "p1")
        protein_2 = Protein(32, 50, "+", "chr1", "p2")
        transcript_1 = Transcript(name="tr1")
        transcript_1.proteins = [protein_1, protein_2]
        self.assertEqual(
            [prot.name for prot in transcript_1.proteins],
            [protein_1.name, protein_2.name]
        )
        self.assertEqual(
            [protein_1.transcript.name, protein_2.transcript.name],
            [transcript_1.name, transcript_1.name]
        )
        # Replace proteins
        protein_1 = Protein(10, 30, "+", "chr1", "p1")
        protein_2 = Protein(32, 50, "+", "chr1", "p2")
        protein_3 = Protein(54, 70, "+", "chr1", "p3")
        transcript_1 = Transcript(name="tr1", proteins=[protein_1, protein_3])
        transcript_1.proteins = [protein_2, protein_3]
        self.assertIsNone(protein_1.transcript)
        self.assertNotIn(protein_1, transcript_1.proteins)
        self.assertEqual(protein_2.transcript, transcript_1)
        self.assertIn(protein_2, transcript_1.proteins)
        self.assertEqual(protein_3.transcript, transcript_1)
        self.assertIn(protein_3, transcript_1.proteins)

    def testLength(self):
        self.assertEqual(self.fwd["transcript"].length(), 73)
        self.assertEqual(self.rvs["transcript"].length(), 73)

    def testGetSubFromRegionPos(self):
        # Forward
        self.assertEqual(
            self.fwd["transcript"].getSubFromRegionPos(54),
            (self.fwd["exon_3"], 2)
        )
        # Reverse
        self.assertEqual(
            self.rvs["transcript"].getSubFromRegionPos(73),
            (self.rvs["exon_1"], 21)
        )

    def testGetPosOnRegion(self):
        # Forward
        self.assertEqual(self.fwd["transcript"].getPosOnRegion(10), 1)
        self.assertEqual(self.fwd["transcript"].getPosOnRegion(20), 11)
        self.assertEqual(self.fwd["transcript"].getPosOnRegion(30), 21)
        self.assertEqual(self.fwd["transcript"].getPosOnRegion(40), 22)
        self.assertEqual(self.fwd["transcript"].getPosOnRegion(80), 53)
        # Reverse
        self.assertEqual(self.rvs["transcript"].getPosOnRegion(100), 1)
        self.assertEqual(self.rvs["transcript"].getPosOnRegion(90), 11)
        self.assertEqual(self.rvs["transcript"].getPosOnRegion(80), 21)
        self.assertEqual(self.rvs["transcript"].getPosOnRegion(70), 22)
        self.assertEqual(self.rvs["transcript"].getPosOnRegion(30), 53)
        # Exception out of transcript
        with self.assertRaises(Exception):
            self.fwd["transcript"].getPosOnRegion(8)
        # Exception in intron
        with self.assertRaises(Exception):
            self.fwd["transcript"].getPosOnRegion(31)
        # Exception strand
        unstranded_tr = Transcript(10, 30, None, "chr1")
        with self.assertRaises(Exception):
            unstranded_tr.getPosOnRegion(20)

    def testGetPosOnRef(self):
        # Forward
        self.assertEqual(self.fwd["transcript"].getPosOnRef(1), 10)
        self.assertEqual(self.fwd["transcript"].getPosOnRef(10), 19)
        self.assertEqual(self.fwd["transcript"].getPosOnRef(21), 30)
        self.assertEqual(self.fwd["transcript"].getPosOnRef(22), 40)
        self.assertEqual(self.fwd["transcript"].getPosOnRef(35), 53)
        self.assertEqual(self.fwd["transcript"].getPosOnRef(52), 70)
        self.assertEqual(self.fwd["transcript"].getPosOnRef(53), 80)
        self.assertEqual(self.fwd["transcript"].getPosOnRef(73), 100)
        # Reverse
        self.assertEqual(self.rvs["transcript"].getPosOnRef(1), 100)
        self.assertEqual(self.rvs["transcript"].getPosOnRef(10), 91)
        self.assertEqual(self.rvs["transcript"].getPosOnRef(21), 80)
        self.assertEqual(self.rvs["transcript"].getPosOnRef(22), 70)
        self.assertEqual(self.rvs["transcript"].getPosOnRef(35), 57)
        self.assertEqual(self.rvs["transcript"].getPosOnRef(52), 40)
        self.assertEqual(self.rvs["transcript"].getPosOnRef(53), 30)
        self.assertEqual(self.rvs["transcript"].getPosOnRef(73), 10)

    def testGetSubFromRefPos(self):
        # Forward
        fw_tests = [
            [10, (self.fwd["exon_1"], 1)],
            [20, (self.fwd["exon_1"], 1)],
            [30, (self.fwd["exon_1"], 1)],
            [31, (self.fwd["intron_1"], 1)],
            [35, (self.fwd["intron_1"], 1)],
            [39, (self.fwd["intron_1"], 1)],
            [40, (self.fwd["exon_2"], 2)],
            [50, (self.fwd["exon_2"], 2)],
            [70, (self.fwd["exon_2"], 2)],
            [71, (self.fwd["intron_2"], 2)],
            [75, (self.fwd["intron_2"], 2)],
            [79, (self.fwd["intron_2"], 2)],
            [80, (self.fwd["exon_3"], 3)],
            [90, (self.fwd["exon_3"], 3)],
            [100, (self.fwd["exon_3"], 3)]
        ]
        for param, expected in fw_tests:
            observed = self.fwd["transcript"].getSubFromRefPos(param)
            if isinstance(expected[0], Exon):
                self.assertEqual(observed, expected)
            else:
                self.assertEqual(observed[1], expected[1])
                self.assertEqual(observed[0].getCoordinatesStr(), expected[0].getCoordinatesStr())
            self.assertEqual(observed[0].annot["siblings_idx"], expected[1])  # Test siblings_idx
        # Reverse
        rvs_tests = [
            [10, (self.rvs["exon_1"], 3)],
            [20, (self.rvs["exon_1"], 3)],
            [30, (self.rvs["exon_1"], 3)],
            [31, (self.rvs["intron_1"], 2)],
            [35, (self.rvs["intron_1"], 2)],
            [39, (self.rvs["intron_1"], 2)],
            [40, (self.rvs["exon_2"], 2)],
            [50, (self.rvs["exon_2"], 2)],
            [70, (self.rvs["exon_2"], 2)],
            [71, (self.rvs["intron_2"], 1)],
            [75, (self.rvs["intron_2"], 1)],
            [79, (self.rvs["intron_2"], 1)],
            [80, (self.rvs["exon_3"], 1)],
            [90, (self.rvs["exon_3"], 1)],
            [100, (self.rvs["exon_3"], 1)]
        ]
        for param, expected in rvs_tests:
            observed = self.rvs["transcript"].getSubFromRefPos(param)
            if isinstance(expected[0], Exon):
                self.assertEqual(observed, expected)
            else:
                self.assertEqual(observed[1], expected[1])
                self.assertEqual(observed[0].getCoordinatesStr(), expected[0].getCoordinatesStr())
            self.assertEqual(observed[0].annot["siblings_idx"], expected[1])  # Test siblings_idx


class TestProtein(unittest.TestCase):
    def setUp(self):
        # Forward
        self.fwd = {
            "cds_1": CDS(10, 30, "+", "chr1"),
            "cds_2": CDS(40, 70, "+", "chr1"),
            "cds_3": CDS(80, 99, "+", "chr1")
        }
        self.fwd["protein"] = Protein(
            children=[
                self.fwd["cds_1"], self.fwd["cds_3"], self.fwd["cds_2"]
            ]
        )
        # Reverse
        self.rvs = {
            "cds_1": CDS(10, 30, "-", "chr1"),
            "cds_2": CDS(40, 70, "-", "chr1"),
            "cds_3": CDS(80, 99, "-", "chr1")
        }
        self.rvs["protein"] = Protein(
            children=[
                self.rvs["cds_1"], self.rvs["cds_3"], self.rvs["cds_2"]
            ]
        )

    def testLength(self):
        self.assertEqual(self.fwd["protein"].length(), 72)

    def testAaLength(self):
        self.assertEqual(self.fwd["protein"].aaLength(), 24)

    def testGetPosOnRef(self):
        # Forward
        self.assertEqual(self.fwd["protein"].getPosOnRef(1, 1), 10)
        self.assertEqual(self.fwd["protein"].getPosOnRef(1, 3), 12)
        self.assertEqual(self.fwd["protein"].getPosOnRef(2, 1), 13)
        self.assertEqual(self.fwd["protein"].getPosOnRef(7, 3), 30)
        self.assertEqual(self.fwd["protein"].getPosOnRef(8, 1), 40)
        self.assertEqual(self.fwd["protein"].getPosOnRef(18, 1), 70)
        self.assertEqual(self.fwd["protein"].getPosOnRef(18, 2), 80)
        self.assertEqual(self.fwd["protein"].getPosOnRef(24, 1), 97)
        self.assertEqual(self.fwd["protein"].getPosOnRef(24, 3), 99)
        # Reverse
        self.assertEqual(self.rvs["protein"].getPosOnRef(1, 1), 99)
        self.assertEqual(self.rvs["protein"].getPosOnRef(1, 3), 97)
        self.assertEqual(self.rvs["protein"].getPosOnRef(2, 1), 96)
        self.assertEqual(self.rvs["protein"].getPosOnRef(7, 2), 80)
        self.assertEqual(self.rvs["protein"].getPosOnRef(7, 3), 70)
        self.assertEqual(self.rvs["protein"].getPosOnRef(17, 3), 40)
        self.assertEqual(self.rvs["protein"].getPosOnRef(18, 1), 30)
        self.assertEqual(self.rvs["protein"].getPosOnRef(24, 1), 12)
        self.assertEqual(self.rvs["protein"].getPosOnRef(24, 3), 10)

    def testGetPosOnRegion(self):
        # Forward
        self.assertEqual(self.fwd["protein"].getPosOnRegion(10), (1, 1))
        self.assertEqual(self.fwd["protein"].getPosOnRegion(12), (1, 3))
        self.assertEqual(self.fwd["protein"].getPosOnRegion(13), (2, 1))
        self.assertEqual(self.fwd["protein"].getPosOnRegion(30), (7, 3))
        self.assertEqual(self.fwd["protein"].getPosOnRegion(40), (8, 1))
        self.assertEqual(self.fwd["protein"].getPosOnRegion(70), (18, 1))
        self.assertEqual(self.fwd["protein"].getPosOnRegion(80), (18, 2))
        self.assertEqual(self.fwd["protein"].getPosOnRegion(97), (24, 1))
        self.assertEqual(self.fwd["protein"].getPosOnRegion(99), (24, 3))
        # Reverse
        self.assertEqual(self.rvs["protein"].getPosOnRegion(99), (1, 1))
        self.assertEqual(self.rvs["protein"].getPosOnRegion(97), (1, 3))
        self.assertEqual(self.rvs["protein"].getPosOnRegion(96), (2, 1))
        self.assertEqual(self.rvs["protein"].getPosOnRegion(80), (7, 2))
        self.assertEqual(self.rvs["protein"].getPosOnRegion(70), (7, 3))
        self.assertEqual(self.rvs["protein"].getPosOnRegion(40), (17, 3))
        self.assertEqual(self.rvs["protein"].getPosOnRegion(30), (18, 1))
        self.assertEqual(self.rvs["protein"].getPosOnRegion(12), (24, 1))
        self.assertEqual(self.rvs["protein"].getPosOnRegion(10), (24, 3))

    def testGetNtPosFromRegionPos(self):
        # Forward
        self.assertEqual(self.fwd["protein"].getNtPosFromRegionPos(1, 1), 1)
        self.assertEqual(self.fwd["protein"].getNtPosFromRegionPos(1, 3), 3)
        self.assertEqual(self.fwd["protein"].getNtPosFromRegionPos(2, 1), 4)
        self.assertEqual(self.fwd["protein"].getNtPosFromRegionPos(7, 3), 21)
        self.assertEqual(self.fwd["protein"].getNtPosFromRegionPos(8, 1), 22)
        self.assertEqual(self.fwd["protein"].getNtPosFromRegionPos(18, 1), 52)
        self.assertEqual(self.fwd["protein"].getNtPosFromRegionPos(18, 2), 53)
        self.assertEqual(self.fwd["protein"].getNtPosFromRegionPos(24, 1), 70)
        self.assertEqual(self.fwd["protein"].getNtPosFromRegionPos(24, 3), 72)
        # Reverse
        self.assertEqual(self.rvs["protein"].getNtPosFromRegionPos(1, 1), 1)
        self.assertEqual(self.rvs["protein"].getNtPosFromRegionPos(1, 3), 3)
        self.assertEqual(self.rvs["protein"].getNtPosFromRegionPos(2, 1), 4)
        self.assertEqual(self.rvs["protein"].getNtPosFromRegionPos(7, 2), 20)
        self.assertEqual(self.rvs["protein"].getNtPosFromRegionPos(7, 3), 21)
        self.assertEqual(self.rvs["protein"].getNtPosFromRegionPos(17, 3), 51)
        self.assertEqual(self.rvs["protein"].getNtPosFromRegionPos(18, 1), 52)
        self.assertEqual(self.rvs["protein"].getNtPosFromRegionPos(24, 1), 70)
        self.assertEqual(self.rvs["protein"].getNtPosFromRegionPos(24, 3), 72)

    def testGetNtPosFromRefPos(self):
        # Forward
        self.assertEqual(self.fwd["protein"].getNtPosFromRefPos(10), 1)
        self.assertEqual(self.fwd["protein"].getNtPosFromRefPos(12), 3)
        self.assertEqual(self.fwd["protein"].getNtPosFromRefPos(13), 4)
        self.assertEqual(self.fwd["protein"].getNtPosFromRefPos(30), 21)
        self.assertEqual(self.fwd["protein"].getNtPosFromRefPos(40), 22)
        self.assertEqual(self.fwd["protein"].getNtPosFromRefPos(70), 52)
        self.assertEqual(self.fwd["protein"].getNtPosFromRefPos(80), 53)
        self.assertEqual(self.fwd["protein"].getNtPosFromRefPos(97), 70)
        self.assertEqual(self.fwd["protein"].getNtPosFromRefPos(99), 72)
        # Reverse
        self.assertEqual(self.rvs["protein"].getNtPosFromRefPos(99), 1)
        self.assertEqual(self.rvs["protein"].getNtPosFromRefPos(97), 3)
        self.assertEqual(self.rvs["protein"].getNtPosFromRefPos(96), 4)
        self.assertEqual(self.rvs["protein"].getNtPosFromRefPos(80), 20)
        self.assertEqual(self.rvs["protein"].getNtPosFromRefPos(70), 21)
        self.assertEqual(self.rvs["protein"].getNtPosFromRefPos(40), 51)
        self.assertEqual(self.rvs["protein"].getNtPosFromRefPos(30), 52)
        self.assertEqual(self.rvs["protein"].getNtPosFromRefPos(12), 70)
        self.assertEqual(self.rvs["protein"].getNtPosFromRefPos(10), 72)

    def testGetSubFromRegionPos(self):
        # Forward
        self.assertEqual(
            self.fwd["protein"].getSubFromRegionPos(1, 1),
            (self.fwd["cds_1"], 1)
        )
        self.assertEqual(
            self.fwd["protein"].getSubFromRegionPos(1, 3),
            (self.fwd["cds_1"], 3)
        )
        self.assertEqual(
            self.fwd["protein"].getSubFromRegionPos(2, 1),
            (self.fwd["cds_1"], 4)
        )
        self.assertEqual(
            self.fwd["protein"].getSubFromRegionPos(7, 3),
            (self.fwd["cds_1"], 21)
        )
        self.assertEqual(
            self.fwd["protein"].getSubFromRegionPos(8, 1),
            (self.fwd["cds_2"], 1)
        )
        self.assertEqual(
            self.fwd["protein"].getSubFromRegionPos(18, 1),
            (self.fwd["cds_2"], 31)
        )
        self.assertEqual(
            self.fwd["protein"].getSubFromRegionPos(18, 2),
            (self.fwd["cds_3"], 1)
        )
        self.assertEqual(
            self.fwd["protein"].getSubFromRegionPos(24, 1),
            (self.fwd["cds_3"], 18)
        )
        self.assertEqual(
            self.fwd["protein"].getSubFromRegionPos(24, 3),
            (self.fwd["cds_3"], 20)
        )
        # Reverse
        self.assertEqual(
            self.rvs["protein"].getSubFromRegionPos(1, 1),
            (self.rvs["cds_3"], 1)
        )
        self.assertEqual(
            self.rvs["protein"].getSubFromRegionPos(1, 3),
            (self.rvs["cds_3"], 3)
        )
        self.assertEqual(
            self.rvs["protein"].getSubFromRegionPos(2, 1),
            (self.rvs["cds_3"], 4)
        )
        self.assertEqual(
            self.rvs["protein"].getSubFromRegionPos(7, 2),
            (self.rvs["cds_3"], 20)
        )
        self.assertEqual(
            self.rvs["protein"].getSubFromRegionPos(7, 3),
            (self.rvs["cds_2"], 1)
        )
        self.assertEqual(
            self.rvs["protein"].getSubFromRegionPos(17, 3),
            (self.rvs["cds_2"], 31)
        )
        self.assertEqual(
            self.rvs["protein"].getSubFromRegionPos(18, 1),
            (self.rvs["cds_1"], 1)
        )
        self.assertEqual(
            self.rvs["protein"].getSubFromRegionPos(24, 1),
            (self.rvs["cds_1"], 19)
        )
        self.assertEqual(
            self.rvs["protein"].getSubFromRegionPos(24, 3),
            (self.rvs["cds_1"], 21)
        )

    def testHasOverlap(self):
        # Forward
        self.assertEqual(self.fwd["protein"].hasOverlap(Region(8, 9, "+", "chr1")), False)
        self.assertEqual(self.fwd["protein"].hasOverlap(Region(9, 9, "+", "chr1")), False)
        self.assertEqual(self.fwd["protein"].hasOverlap(Region(9, 10, "+", "chr1")), True)
        self.assertEqual(self.fwd["protein"].hasOverlap(Region(10, 10, "+", "chr1")), True)
        self.assertEqual(self.fwd["protein"].hasOverlap(Region(12, 12, "+", "chr1")), True)
        self.assertEqual(self.fwd["protein"].hasOverlap(Region(12, 29, "+", "chr1")), True)
        self.assertEqual(self.fwd["protein"].hasOverlap(Region(30, 30, "+", "chr1")), True)
        self.assertEqual(self.fwd["protein"].hasOverlap(Region(30, 31, "+", "chr1")), True)
        self.assertEqual(self.fwd["protein"].hasOverlap(Region(31, 31, "+", "chr1")), False)
        self.assertEqual(self.fwd["protein"].hasOverlap(Region(37, 39, "+", "chr1")), False)
        self.assertEqual(self.fwd["protein"].hasOverlap(Region(39, 39, "+", "chr1")), False)
        self.assertEqual(self.fwd["protein"].hasOverlap(Region(39, 40, "+", "chr1")), True)
        self.assertEqual(self.fwd["protein"].hasOverlap(Region(40, 40, "+", "chr1")), True)
        self.assertEqual(self.fwd["protein"].hasOverlap(Region(98, 99, "+", "chr1")), True)
        self.assertEqual(self.fwd["protein"].hasOverlap(Region(100, 102, "+", "chr1")), False)
        # Reverse
        self.assertEqual(self.rvs["protein"].hasOverlap(Region(8, 9, "-", "chr1")), False)
        self.assertEqual(self.rvs["protein"].hasOverlap(Region(9, 9, "-", "chr1")), False)
        self.assertEqual(self.rvs["protein"].hasOverlap(Region(9, 10, "-", "chr1")), True)
        self.assertEqual(self.rvs["protein"].hasOverlap(Region(10, 10, "-", "chr1")), True)
        self.assertEqual(self.rvs["protein"].hasOverlap(Region(12, 12, "-", "chr1")), True)
        self.assertEqual(self.rvs["protein"].hasOverlap(Region(12, 29, "-", "chr1")), True)
        self.assertEqual(self.rvs["protein"].hasOverlap(Region(30, 30, "-", "chr1")), True)
        self.assertEqual(self.rvs["protein"].hasOverlap(Region(30, 31, "-", "chr1")), True)
        self.assertEqual(self.rvs["protein"].hasOverlap(Region(31, 31, "-", "chr1")), False)
        self.assertEqual(self.rvs["protein"].hasOverlap(Region(37, 39, "-", "chr1")), False)
        self.assertEqual(self.rvs["protein"].hasOverlap(Region(39, 39, "-", "chr1")), False)
        self.assertEqual(self.rvs["protein"].hasOverlap(Region(39, 40, "-", "chr1")), True)
        self.assertEqual(self.rvs["protein"].hasOverlap(Region(40, 40, "-", "chr1")), True)
        self.assertEqual(self.rvs["protein"].hasOverlap(Region(98, 99, "-", "chr1")), True)
        self.assertEqual(self.rvs["protein"].hasOverlap(Region(100, 102, "-", "chr1")), False)

    def testGetCDSFromTranscript(self):
        res = []
        # One exon forward
        transcript_1 = Transcript(None, None, "+", "chr1", children=[
            Exon(100, 150, "+", "chr1")
        ])
        res.append({
            "expected": [CDS(100, 150, "+", "chr1")],
            "observed": Protein(100, 150, "+", "chr1", transcript=transcript_1).getCDSFromTranscript()
        })
        res.append({
            "expected": [CDS(110, 140, "+", "chr1")],
            "observed": Protein(110, 140, "+", "chr1", transcript=transcript_1).getCDSFromTranscript()
        })
        res.append({
            "expected": [CDS(110, 150, "+", "chr1")],
            "observed": Protein(110, 150, "+", "chr1", transcript=transcript_1).getCDSFromTranscript()
        })
        res.append({
            "expected": [CDS(100, 140, "+", "chr1")],
            "observed": Protein(100, 140, "+", "chr1", transcript=transcript_1).getCDSFromTranscript()
        })
        # Three exons forward
        transcript_2 = Transcript(None, None, "+", "chr1", children=[
            Exon(30, 80, "+", "chr1"),
            Exon(100, 150, "+", "chr1"),
            Exon(170, 200, "+", "chr1")
        ])
        res.append({
            "expected": [CDS(100, 150, "+", "chr1")],
            "observed": Protein(100, 150, "+", "chr1", transcript=transcript_2).getCDSFromTranscript()
        })
        res.append({
            "expected": [CDS(110, 140, "+", "chr1")],
            "observed": Protein(110, 140, "+", "chr1", transcript=transcript_2).getCDSFromTranscript()
        })
        res.append({
            "expected": [CDS(110, 150, "+", "chr1")],
            "observed": Protein(110, 150, "+", "chr1", transcript=transcript_2).getCDSFromTranscript()
        })
        res.append({
            "expected": [CDS(100, 140, "+", "chr1")],
            "observed": Protein(100, 140, "+", "chr1", transcript=transcript_2).getCDSFromTranscript()
        })
        res.append({
            "expected": [CDS(30, 80, "+", "chr1"), CDS(100, 150, "+", "chr1")],
            "observed": Protein(30, 150, "+", "chr1", transcript=transcript_2).getCDSFromTranscript()
        })
        res.append({
            "expected": [CDS(60, 80, "+", "chr1"), CDS(100, 150, "+", "chr1")],
            "observed": Protein(60, 150, "+", "chr1", transcript=transcript_2).getCDSFromTranscript()
        })
        res.append({
            "expected": [CDS(80, 80, "+", "chr1"), CDS(100, 150, "+", "chr1")],
            "observed": Protein(80, 150, "+", "chr1", transcript=transcript_2).getCDSFromTranscript()
        })
        res.append({
            "expected": [CDS(80, 80, "+", "chr1"), CDS(100, 150, "+", "chr1"), CDS(170, 170, "+", "chr1")],
            "observed": Protein(80, 170, "+", "chr1", transcript=transcript_2).getCDSFromTranscript()
        })
        res.append({
            "expected": [CDS(100, 150, "+", "chr1"), CDS(170, 200, "+", "chr1")],
            "observed": Protein(100, 200, "+", "chr1", transcript=transcript_2).getCDSFromTranscript()
        })
        res.append({
            "expected": [CDS(100, 150, "+", "chr1"), CDS(170, 190, "+", "chr1")],
            "observed": Protein(100, 190, "+", "chr1", transcript=transcript_2).getCDSFromTranscript()
        })
        res.append({
            "expected": [CDS(110, 150, "+", "chr1"), CDS(170, 200, "+", "chr1")],
            "observed": Protein(110, 200, "+", "chr1", transcript=transcript_2).getCDSFromTranscript()
        })
        res.append({
            "expected": [CDS(110, 150, "+", "chr1"), CDS(170, 190, "+", "chr1")],
            "observed": Protein(110, 190, "+", "chr1", transcript=transcript_2).getCDSFromTranscript()
        })
        # One exon reverse
        transcript_1 = Transcript(None, None, "-", "chr1", children=[
            Exon(100, 150, "-", "chr1")
        ])
        res.append({
            "expected": [CDS(100, 150, "-", "chr1")],
            "observed": Protein(100, 150, "-", "chr1", transcript=transcript_1).getCDSFromTranscript()
        })
        res.append({
            "expected": [CDS(110, 140, "-", "chr1")],
            "observed": Protein(110, 140, "-", "chr1", transcript=transcript_1).getCDSFromTranscript()
        })
        res.append({
            "expected": [CDS(110, 150, "-", "chr1")],
            "observed": Protein(110, 150, "-", "chr1", transcript=transcript_1).getCDSFromTranscript()
        })
        res.append({
            "expected": [CDS(100, 140, "-", "chr1")],
            "observed": Protein(100, 140, "-", "chr1", transcript=transcript_1).getCDSFromTranscript()
        })
        # Three exons reverse
        transcript_2 = Transcript(None, None, "-", "chr1", children=[
            Exon(170, 200, "-", "chr1"),
            Exon(100, 150, "-", "chr1"),
            Exon(30, 80, "-", "chr1")
        ])
        res.append({
            "expected": [CDS(100, 150, "-", "chr1")],
            "observed": Protein(100, 150, "-", "chr1", transcript=transcript_2).getCDSFromTranscript()
        })
        res.append({
            "expected": [CDS(110, 140, "-", "chr1")],
            "observed": Protein(110, 140, "-", "chr1", transcript=transcript_2).getCDSFromTranscript()
        })
        res.append({
            "expected": [CDS(110, 150, "-", "chr1")],
            "observed": Protein(110, 150, "-", "chr1", transcript=transcript_2).getCDSFromTranscript()
        })
        res.append({
            "expected": [CDS(100, 140, "-", "chr1")],
            "observed": Protein(100, 140, "-", "chr1", transcript=transcript_2).getCDSFromTranscript()
        })
        res.append({
            "expected": [CDS(100, 150, "-", "chr1"), CDS(30, 80, "-", "chr1")],
            "observed": Protein(30, 150, "-", "chr1", transcript=transcript_2).getCDSFromTranscript()
        })
        res.append({
            "expected": [CDS(100, 150, "-", "chr1"), CDS(60, 80, "-", "chr1")],
            "observed": Protein(60, 150, "-", "chr1", transcript=transcript_2).getCDSFromTranscript()
        })
        res.append({
            "expected": [CDS(100, 150, "-", "chr1"), CDS(80, 80, "-", "chr1")],
            "observed": Protein(80, 150, "-", "chr1", transcript=transcript_2).getCDSFromTranscript()
        })
        res.append({
            "expected": [CDS(170, 170, "-", "chr1"), CDS(100, 150, "-", "chr1"), CDS(80, 80, "-", "chr1")],
            "observed": Protein(80, 170, "-", "chr1", transcript=transcript_2).getCDSFromTranscript()
        })
        res.append({
            "expected": [CDS(170, 200, "-", "chr1"), CDS(100, 150, "-", "chr1")],
            "observed": Protein(100, 200, "-", "chr1", transcript=transcript_2).getCDSFromTranscript()
        })
        res.append({
            "expected": [CDS(170, 190, "-", "chr1"), CDS(100, 150, "-", "chr1")],
            "observed": Protein(100, 190, "-", "chr1", transcript=transcript_2).getCDSFromTranscript()
        })
        res.append({
            "expected": [CDS(170, 200, "-", "chr1"), CDS(110, 150, "-", "chr1")],
            "observed": Protein(110, 200, "-", "chr1", transcript=transcript_2).getCDSFromTranscript()
        })
        res.append({
            "expected": [CDS(170, 190, "-", "chr1"), CDS(110, 150, "-", "chr1")],
            "observed": Protein(110, 190, "-", "chr1", transcript=transcript_2).getCDSFromTranscript()
        })
        # Launch evaluation
        for eval_pair in res:
            self.assertEqual(
                ", ".join([curr_cds.getCoordinatesStr() for curr_cds in eval_pair["expected"]]),
                ", ".join([curr_cds.getCoordinatesStr() for curr_cds in eval_pair["observed"]]),
            )

    def testGetCodonRefPos(self):
        tr_1 = Transcript(None, None, "+", "chr1", children=[
            Exon(10, 20, "+", "chr1"),
            Exon(30, 40, "+", "chr1"),
            Exon(45, 50, "+", "chr1")
        ])
        prot_1 = Protein(16, 36, "+", "chr1", transcript=tr_1)
        tr_2 = Transcript(None, None, "-", "chr1", children=[
            Exon(10, 20, "-", "chr1"),
            Exon(30, 40, "-", "chr1"),
            Exon(45, 50, "-", "chr1")
        ])
        prot_2 = Protein(16, 36, "-", "chr1", transcript=tr_2)
        data = [
            {"prot": prot_1, "aa_pos": 1, "expected": [16, 17, 18]},
            {"prot": prot_1, "aa_pos": 2, "expected": [19, 20, 30]},
            {"prot": prot_1, "aa_pos": 3, "expected": [31, 32, 33]},
            {"prot": prot_1, "aa_pos": 4, "expected": [34, 35, 36]},
            {"prot": prot_2, "aa_pos": 1, "expected": [36, 35, 34]},
            {"prot": prot_2, "aa_pos": 2, "expected": [33, 32, 31]},
            {"prot": prot_2, "aa_pos": 3, "expected": [30, 20, 19]},
            {"prot": prot_2, "aa_pos": 4, "expected": [18, 17, 16]},
        ]
        for curr in data:
            self.assertEqual(
                curr["prot"].getCodonRefPos(curr["aa_pos"]),
                curr["expected"]
            )
            with self.assertRaises(Exception):
                self.prot_1.getCodonRefPos(8)  # Not in protein


class TestProteinSeq(unittest.TestCase):
    def setUp(self):
        tmp_folder = tempfile.gettempdir()
        unique_id = str(uuid.uuid1())

        # Temporary files
        self.tmp_fasta_idx = os.path.join(tmp_folder, unique_id + ".fasta.fai")
        self.tmp_fasta = os.path.join(tmp_folder, unique_id + ".fasta")

        # Create sequence file
        content_fasta = """>one
ATGCATGCATGCATGCATGCATGCATGCAT
GCATGCATGCATGCATGCATGCATGCATGC
ATGCAT
>two another chromosome
ATGCATGCATGCAT
GCATGCATGCATGC"""
        with open(self.tmp_fasta, "w") as FH_out:
            FH_out.write(content_fasta)

        # Proteins
        # 1 3 5 7 9  11 14 16 19 21        30 33 36 39 41  44 47 50
        # ATGCATGCAT GCATG CATGC ATGCATGCA TGCATGCATGC ATGCATGCATGCATGCATGC
        #            ..... *****           *******....      ......
        # prot_1           12345           6789 11
        tr_1 = Transcript(None, None, "+", "one", children=[
            Exon(10, 20, "+", "one"),
            Exon(30, 40, "+", "one"),
            Exon(45, 50, "+", "one")
        ])
        self.prot_1 = Protein(16, 36, "+", "one", transcript=tr_1)
        tr_2 = Transcript(None, None, "-", "one", children=[
            Exon(10, 20, "-", "one"),
            Exon(30, 40, "-", "one"),
            Exon(45, 50, "-", "one")
        ])
        self.prot_2 = Protein(16, 36, "-", "one", transcript=tr_2)

        # Create index
        content_fasta_idx = """one	66	5	30	31
two	28	98	14	15"""
        with open(self.tmp_fasta_idx, "w") as FH_out:
            FH_out.write(content_fasta_idx)

    def tearDown(self):
        # Clean temporary files
        for curr_file in [self.tmp_fasta, self.tmp_fasta_idx]:
            if os.path.exists(curr_file):
                os.remove(curr_file)

    def testGetCodonSeqFromProtPos(self):
        data = [
            {"prot": self.prot_1, "aa_pos": 1, "expected": "CAT"},
            {"prot": self.prot_1, "aa_pos": 2, "expected": "GCT"},
            {"prot": self.prot_1, "aa_pos": 4, "expected": "TGC"},
            {"prot": self.prot_2, "aa_pos": 1, "expected": "GCA"},
            {"prot": self.prot_2, "aa_pos": 3, "expected": "AGC"},
            {"prot": self.prot_2, "aa_pos": 4, "expected": "ATG"},
        ]
        with IdxFastaIO(self.tmp_fasta) as reader:
            for curr in data:
                self.assertEqual(
                    curr["prot"].getCodonSeqFromProtPos(curr["aa_pos"], reader),
                    curr["expected"]
                )
            with self.assertRaises(Exception):
                self.prot_1.getCodonSeqFromProtPos(8, reader)  # Not in protein

    def testGetCodonInfo(self):
        data = [
            {"prot": self.prot_1, "ref_pos": 16, "expected": (1, 1, "CAT")},
            {"prot": self.prot_1, "ref_pos": 18, "expected": (1, 3, "CAT")},
            {"prot": self.prot_1, "ref_pos": 19, "expected": (2, 1, "GCT")},
            {"prot": self.prot_1, "ref_pos": 20, "expected": (2, 2, "GCT")},
            {"prot": self.prot_1, "ref_pos": 34, "expected": (4, 1, "TGC")},
            {"prot": self.prot_1, "ref_pos": 35, "expected": (4, 2, "TGC")},
            {"prot": self.prot_1, "ref_pos": 36, "expected": (4, 3, "TGC")},
            {"prot": self.prot_2, "ref_pos": 36, "expected": (1, 1, "GCA")},
            {"prot": self.prot_2, "ref_pos": 34, "expected": (1, 3, "GCA")},
            {"prot": self.prot_2, "ref_pos": 30, "expected": (3, 1, "AGC")},
            {"prot": self.prot_2, "ref_pos": 20, "expected": (3, 2, "AGC")},
            {"prot": self.prot_2, "ref_pos": 19, "expected": (3, 3, "AGC")},
            {"prot": self.prot_2, "ref_pos": 18, "expected": (4, 1, "ATG")},
            {"prot": self.prot_2, "ref_pos": 16, "expected": (4, 3, "ATG")}
        ]
        with IdxFastaIO(self.tmp_fasta) as reader:
            for curr in data:
                self.assertEqual(
                    curr["prot"].getCodonInfo(curr["ref_pos"], reader),
                    curr["expected"]
                )
            with self.assertRaises(Exception):
                self.prot_1.getCodonInfo(14, reader)  # In exon but not in CDS


########################################################################
#
# MAIN
#
########################################################################
if __name__ == "__main__":
    unittest.main()
